@@ -10,6 +10,7 @@
  * Split from fs/nfs/super.c by David Howells <dhowells@redhat.com>
  */
 
+#include <linux/compat.h>
 #include <linux/module.h>
 #include <linux/fs.h>
 #include <linux/fs_context.h>
@@ -20,10 +21,17 @@
 #include <linux/inetdevice.h>
 #include <linux/compat.h>
 #include <linux/dns_resolver.h>
+#include <linux/vmalloc.h>
+
+#include <net/handshake.h>
+#include <net/addrconf.h>
+
 #include "nfs.h"
 #include "nfstrace.h"
 #include "internal.h"
 
+#include "nfstrace.h"
+
 #define NFSDBG_FACILITY		NFSDBG_MOUNT
 
 #if IS_ENABLED(CONFIG_NFS_V3)
@@ -72,7 +80,6 @@ enum nfs_param {
 	Opt_mountvers,
 	Opt_namelen,
 	Opt_nconnect,
-	Opt_pconnect,
 	Opt_max_connect,
 	Opt_localports,
 	Opt_localports_failover,
@@ -84,6 +91,8 @@ enum nfs_param {
 	Opt_posix,
 	Opt_proto,
 	Opt_rdirplus,
+	Opt_rdirplus_none,
+	Opt_rdirplus_force,
 	Opt_rdma,
 	Opt_resvport,
 	Opt_retrans,
@@ -101,11 +110,13 @@ enum nfs_param {
 	Opt_source,
 	Opt_tcp,
 	Opt_timeo,
+	Opt_trunkdiscovery,
 	Opt_udp,
 	Opt_v,
 	Opt_vers,
 	Opt_wsize,
 	Opt_write,
+	Opt_xprtsec,
 };
 
 enum {
@@ -115,7 +126,6 @@ enum {
 	Opt_local_lock_posix,
 };
 
-#if !defined(COMPAT_DETECT_FS_PARAM_DESCRIPTION_ENUMS)
 static const struct constant_table nfs_param_enums_local_lock[] = {
 	{ "all",		Opt_local_lock_all },
 	{ "flock",	Opt_local_lock_flock },
@@ -123,7 +133,6 @@ static const struct constant_table nfs_param_enums_local_lock[] = {
 	{ "none",		Opt_local_lock_none },
 	{}
 };
-#endif
 
 enum {
 	Opt_lookupcache_all,
@@ -131,7 +140,6 @@ enum {
 	Opt_lookupcache_positive,
 };
 
-#if !defined(COMPAT_DETECT_FS_PARAM_DESCRIPTION_ENUMS)
 static const struct constant_table nfs_param_enums_lookupcache[] = {
 	{ "all",		Opt_lookupcache_all },
 	{ "none",		Opt_lookupcache_none },
@@ -139,7 +147,6 @@ static const struct constant_table nfs_param_enums_lookupcache[] = {
 	{ "positive",		Opt_lookupcache_positive },
 	{}
 };
-#endif
 
 enum {
 	Opt_write_lazy,
@@ -147,23 +154,6 @@ enum {
 	Opt_write_wait,
 };
 
-#if defined(COMPAT_DETECT_FS_PARAM_DESCRIPTION_ENUMS)
-static const struct fs_parameter_enum nfs_param_enums[] = {
-       { Opt_local_lock,       "all",          Opt_local_lock_all },
-       { Opt_local_lock,       "flock",        Opt_local_lock_flock },
-       { Opt_local_lock,       "none",         Opt_local_lock_none },
-       { Opt_local_lock,       "posix",        Opt_local_lock_posix },
-       { Opt_lookupcache,      "all",          Opt_lookupcache_all },
-       { Opt_lookupcache,      "none",         Opt_lookupcache_none },
-       { Opt_lookupcache,      "pos",          Opt_lookupcache_positive },
-       { Opt_lookupcache,      "positive",     Opt_lookupcache_positive },
-       { Opt_write,	       "lazy",	       Opt_write_lazy },
-       { Opt_write,	       "eager",        Opt_write_eager },
-       { Opt_write,	       "wait",         Opt_write_wait },
-       {}
-};
-#endif
-
 static const struct constant_table nfs_param_enums_write[] = {
 	{ "lazy",		Opt_write_lazy },
 	{ "eager",		Opt_write_eager },
@@ -171,12 +161,7 @@ static const struct constant_table nfs_param_enums_write[] = {
 	{}
 };
 
-#if defined(COMPAT_DETECT_FS_PARAM_DESCRIPTION)
-static const struct fs_parameter_spec nfs_param_specs[]
-#else
-static const struct fs_parameter_spec nfs_fs_parameters[]
-#endif
- = {
+static const struct fs_parameter_spec nfs_fs_parameters[] = {
 	fsparam_flag_no("ac",		Opt_ac),
 	fsparam_u32   ("acdirmax",	Opt_acdirmax),
 	fsparam_u32   ("acdirmin",	Opt_acdirmin),
@@ -191,19 +176,14 @@ static const struct fs_parameter_spec nfs_fs_parameters[]
 	fsparam_flag_no("cto",		Opt_cto),
 	fsparam_flag_no("extend",	Opt_extend),
 	fsparam_flag  ("fg",		Opt_fg),
-#ifdef COMPAT_DETECT_FS_PARAM_V_OPTIONAL
-	__fsparam_compat(fs_param_is_string, "fsc",            Opt_fscache,
-			  fs_param_neg_with_no|fs_param_v_optional, NULL),
-#else
 	fsparam_flag_no("fsc",		Opt_fscache_flag),
 	fsparam_string("fsc",		Opt_fscache),
-#endif
 	fsparam_flag  ("hard",		Opt_hard),
-	__fsparam_compat(fsparam_type_is_flag, "intr",		Opt_intr,
+	__fsparam(NULL, "intr",		Opt_intr,
 		  fs_param_neg_with_no|fs_param_deprecated, NULL),
-	fsparam_enum_compat  ("local_lock",	Opt_local_lock, nfs_param_enums_local_lock),
+	fsparam_enum  ("local_lock",	Opt_local_lock, nfs_param_enums_local_lock),
 	fsparam_flag_no("lock",		Opt_lock),
-	fsparam_enum_compat  ("lookupcache",	Opt_lookupcache, nfs_param_enums_lookupcache),
+	fsparam_enum  ("lookupcache",	Opt_lookupcache, nfs_param_enums_lookupcache),
 	fsparam_flag_no("migration",	Opt_migration),
 	fsparam_flag_no("idlexprt",	Opt_idlexprt),
 	fsparam_u32   ("mdconnect",	Opt_mdconnect),
@@ -224,13 +204,13 @@ static const struct fs_parameter_spec nfs_fs_parameters[]
 	fsparam_flag_no("spread_reads",	Opt_spread_reads),
 	fsparam_flag_no("spread_writes", Opt_spread_writes),
 	fsparam_string("nfsvers",	Opt_vers),
-	fsparam_u32   ("pconnect",	Opt_pconnect),
 	fsparam_u32   ("port",		Opt_port),
 	fsparam_flag_no("posix",	Opt_posix),
 	fsparam_string("proto",		Opt_proto),
 	fsparam_flag_no("forcerdirplus",	Opt_forcerdirplus),
 	fsparam_flag  ("relmtime",  Opt_relmtime),
-	fsparam_flag_no("rdirplus",	Opt_rdirplus),
+	fsparam_flag_no("rdirplus", Opt_rdirplus), // rdirplus|nordirplus
+	fsparam_string("rdirplus",  Opt_rdirplus), // rdirplus=...
 	fsparam_flag  ("rdma",		Opt_rdma),
 	fsparam_flag_no("resvport",	Opt_resvport),
 	fsparam_u32   ("retrans",	Opt_retrans),
@@ -247,6 +227,7 @@ static const struct fs_parameter_spec nfs_fs_parameters[]
 	fsparam_string("source",	Opt_source),
 	fsparam_flag  ("tcp",		Opt_tcp),
 	fsparam_u32   ("timeo",		Opt_timeo),
+	fsparam_flag_no("trunkdiscovery", Opt_trunkdiscovery),
 	fsparam_flag  ("udp",		Opt_udp),
 	fsparam_flag  ("v2",		Opt_v),
 	fsparam_flag  ("v3",		Opt_v),
@@ -255,24 +236,12 @@ static const struct fs_parameter_spec nfs_fs_parameters[]
 	fsparam_flag  ("v4.1",		Opt_v),
 	fsparam_flag  ("v4.2",		Opt_v),
 	fsparam_string("vers",		Opt_vers),
-	fsparam_enum_compat  ("write",		Opt_write, nfs_param_enums_write),
+	fsparam_enum  ("write",		Opt_write, nfs_param_enums_write),
 	fsparam_u32   ("wsize",		Opt_wsize),
+	fsparam_string("xprtsec",	Opt_xprtsec),
 	{}
 };
 
-#if defined(COMPAT_DETECT_FS_PARAM_DESCRIPTION)
-static const struct fs_parameter_description nfs_fs_parameters = {
-       .name           = "nfs",
-       .specs          = nfs_param_specs,
-#if defined(COMPAT_DETECT_FS_PARAM_DESCRIPTION_ENUMS)
-       .enums          = nfs_param_enums,
-#endif
-};
-#define NFS_FS_PARAMETERS_REF &nfs_fs_parameters
-#else
-#define NFS_FS_PARAMETERS_REF nfs_fs_parameters
-#endif
-
 enum {
 	Opt_vers_2,
 	Opt_vers_3,
@@ -343,6 +312,20 @@ static const struct constant_table nfs_secflavor_tokens[] = {
 	{}
 };
 
+enum {
+	Opt_xprtsec_none,
+	Opt_xprtsec_tls,
+	Opt_xprtsec_mtls,
+	nr__Opt_xprtsec
+};
+
+static const struct constant_table nfs_xprtsec_policies[] = {
+	{ "none",	Opt_xprtsec_none },
+	{ "tls",	Opt_xprtsec_tls },
+	{ "mtls",	Opt_xprtsec_mtls },
+	{}
+};
+
 static DEFINE_IDA(nfs_fs_context_trids);
 
 void nfs_fs_context_cleanup_trids(void)
@@ -370,8 +353,8 @@ nfs_fs_context_trace(const char *kind, struct fs_context *fc,
 		     const char *fmt, ...)
 {
 	struct nfs_fs_context *ctx = nfs_fc2context(fc);
+	char buf[0x80] = {0, };
 	va_list args;
-	char buf[0x80] = {};
 
 	va_start(args, fmt);
 	vsnprintf(buf, sizeof(buf), fmt, args);
@@ -381,15 +364,21 @@ nfs_fs_context_trace(const char *kind, struct fs_context *fc,
 }
 EXPORT_SYMBOL(nfs_fs_context_trace);
 
+static const struct constant_table nfs_rdirplus_tokens[] = {
+	{ "none",	Opt_rdirplus_none },
+	{ "force",	Opt_rdirplus_force },
+	{}
+};
+
 /*
  * Sanity-check a server address provided by the mount command.
  *
  * Address family must be initialized, and address must not be
  * the ANY address for that family.
  */
-static int nfs_verify_server_address(struct sockaddr *addr)
+static int nfs_verify_server_address(struct sockaddr_storage *addr)
 {
-	switch (addr->sa_family) {
+	switch (addr->ss_family) {
 	case AF_INET: {
 		struct sockaddr_in *sa = (struct sockaddr_in *)addr;
 		return sa->sin_addr.s_addr != htonl(INADDR_ANY);
@@ -400,7 +389,6 @@ static int nfs_verify_server_address(struct sockaddr *addr)
 	}
 	}
 
-	dfprintk(MOUNT, "NFS: Invalid IP address specified\n");
 	return 0;
 }
 
@@ -435,9 +423,21 @@ static int nfs_validate_transport_protocol(struct fs_context *fc,
 	default:
 		ctx->nfs_server.protocol = XPRT_TRANSPORT_TCP;
 	}
+
+	if (ctx->xprtsec.policy != RPC_XPRTSEC_NONE)
+		switch (ctx->nfs_server.protocol) {
+		case XPRT_TRANSPORT_TCP:
+			ctx->nfs_server.protocol = XPRT_TRANSPORT_TCP_TLS;
+			break;
+		default:
+			goto out_invalid_xprtsec_policy;
+	}
+
 	return 0;
 out_invalid_transport_udp:
 	return nfs_invalf(fc, "NFS: Unsupported transport protocol udp");
+out_invalid_xprtsec_policy:
+	return nfs_invalf(fc, "NFS: Transport does not support xprtsec");
 }
 
 /*
@@ -494,7 +494,7 @@ static int nfs_parse_security_flavors(struct fs_context *fc,
 	char *string = param->string, *p;
 	int ret;
 
-	dfprintk(MOUNT, "NFS: parsing %s=%s option\n", param->key, param->string);
+	trace_nfs_mount_assign(param->key, string);
 
 	while ((p = strsep(&string, ":")) != NULL) {
 		if (!*p)
@@ -545,6 +545,29 @@ static int nfs_parse_security_flavors(struct fs_context *fc,
 	return 0;
 }
 
+static int nfs_parse_xprtsec_policy(struct fs_context *fc,
+				    struct fs_parameter *param)
+{
+	struct nfs_fs_context *ctx = nfs_fc2context(fc);
+
+	trace_nfs_mount_assign(param->key, param->string);
+
+	switch (lookup_constant(nfs_xprtsec_policies, param->string, -1)) {
+	case Opt_xprtsec_none:
+		ctx->xprtsec.policy = RPC_XPRTSEC_NONE;
+		break;
+	case Opt_xprtsec_tls:
+		ctx->xprtsec.policy = RPC_XPRTSEC_TLS_ANON;
+		break;
+	case Opt_xprtsec_mtls:
+		ctx->xprtsec.policy = RPC_XPRTSEC_TLS_X509;
+		break;
+	default:
+		return nfs_invalf(fc, "NFS: Unrecognized transport security policy");
+	}
+	return 0;
+}
+
 static int nfs_parse_version_string(struct fs_context *fc,
 				    const char *string)
 {
@@ -591,12 +614,10 @@ static int nfs_parse_netdev_or_address(struct net *net, const char *buf,
 {
 	struct net_device *dev;
 	struct in_device *in_dev;
-	char buf0[IFNAMSIZ] = {};
-	uint32_t ipv4_addr;
-	int res;
+	uint32_t ipv4_addr = 0;
+	int res = rpc_pton(net, buf, buflen, sap, salen);
 
-	res = rpc_pton(net, buf, buflen, sap, salen);
-	if (res)
+	if (res > 0)
 		return res;
 
 	if (strcmp(type, "local"))
@@ -607,25 +628,45 @@ static int nfs_parse_netdev_or_address(struct net *net, const char *buf,
 	if (buflen >= IFNAMSIZ)
 		goto error;
 
-	memcpy(buf0, buf, buflen);
-	dev = dev_get_by_name(net, buf0);
+	dev = dev_get_by_name(net, buf);
 	if (dev == NULL)
 		goto error;
 
 	in_dev = in_dev_get(dev);
 	if (in_dev) {
-		if (!sunrpc_get_in_dev_primary_addr_locked(in_dev, &ipv4_addr)) {
-			/* We found an address */
-			struct sockaddr_in *sin = (struct sockaddr_in *)sap;
-			memset(sin, 0, sizeof(*sin));
-			sin->sin_family = AF_INET;
-			sin->sin_addr.s_addr = ipv4_addr;
-			res = sizeof(struct sockaddr_in);
-		}
+		rcu_read_lock();
+		if (sunrpc_get_in_dev_primary_addr_locked(in_dev, &ipv4_addr))
+			ipv4_addr = 0;
+		rcu_read_unlock();
 		in_dev_put(in_dev);
 	}
-
+	if (ipv4_addr) {
+		/* We found an address */
+		struct sockaddr_in *in = (struct sockaddr_in *)sap;
+		memset(in, 0, sizeof(*in));
+		in->sin_family = AF_INET;
+		in->sin_addr.s_addr = ipv4_addr;
+		res = sizeof(*in);
+	} else {
+		struct inet6_dev *in6_dev = in6_dev_get(dev);
+		if (in6_dev) {
+			const struct inet6_ifaddr *ifp;
+			rcu_read_lock();
+			list_for_each_entry_rcu(ifp, &in6_dev->addr_list, if_list)
+				if (!(ifp->scope & (IPV6_ADDR_LOOPBACK | IPV6_ADDR_MULTICAST))) {
+					struct sockaddr_in6 *in6 = (struct sockaddr_in6 *)sap;
+					memset(in6, 0, sizeof(*in6));
+					in6->sin6_family = AF_INET6;
+					in6->sin6_addr = ifp->addr;
+					res = sizeof(*in6);
+					break;
+				}
+			rcu_read_unlock();
+			in6_dev_put(in6_dev);
+		}
+	}
 	dev_put(dev);
+
 	return res;
 
 error:
@@ -734,25 +775,18 @@ static int nfs_parse_port_group(char *string, struct fs_context *fc,
 
 	string_scan = string;
 	while (string_scan != NULL) {
-		const char *delim;
+		const char *delim = NULL;
 		const char *token = string_scan;
-		const char *a = strchr(string_scan, '~');
-		const char *b = strchr(string_scan, '-');
+		const char *pos = strpbrk(string_scan, "~-");
 		bool single = true;
 
-		delim = NULL;
-		if (b && a) {
-			if (b < a) {
+		if (pos) {
+			if (*pos == '~')
+				delim = strsep(&string_scan, "~");
+			else {
 				single = false;
 				delim = strsep(&string_scan, "-");
-			} else {
-				delim = strsep(&string_scan, "~");
 			}
-		} else if (a) {
-			delim = strsep(&string_scan, "~");
-		} else if (b) {
-			single = false;
-			delim = strsep(&string_scan, "-");
 		}
 
 		if (!prev_range) {
@@ -831,30 +865,15 @@ static int nfs_fs_context_parse_param(struct fs_context *fc,
 	unsigned int len;
 	int ret, opt;
 
-	dfprintk(MOUNT, "NFS:   parsing nfs mount option '%s'\n", param->key);
+	trace_nfs_mount_option(param);
 
-	opt = fs_parse(fc, NFS_FS_PARAMETERS_REF, param, &result);
+	opt = fs_parse(fc, nfs_fs_parameters, param, &result);
 	if (opt < 0)
-		return ctx->sloppy ? 1 : opt;
+		return (opt == -ENOPARAM && ctx->sloppy) ? 1 : opt;
 
-	{
-		char param_str[0x20] = {};
-		const char *value = &param_str[0];
-
-		if (param->type == fs_value_is_string && param->string)
-			value = param->string;
-		else
-			snprintf(param_str, sizeof(param_str),
-				 "%d", result.uint_32);
-
-		trace_nfs_fs_context_parse_param(ctx, param->key, value);
-	}
-
-
-#ifdef COMPAT_DETECT_INCLUDE_FS_CONTEXT
 	if (fc->security)
 		ctx->has_sec_mnt_opts = 1;
-#endif
+
 	switch (opt) {
 	case Opt_source:
 		if (fc->source)
@@ -897,6 +916,12 @@ static int nfs_fs_context_parse_param(struct fs_context *fc,
 		else
 			ctx->flags &= ~NFS_MOUNT_NOCTO;
 		break;
+	case Opt_trunkdiscovery:
+		if (result.negated)
+			ctx->flags &= ~NFS_MOUNT_TRUNK_DISCOVERY;
+		else
+			ctx->flags |= NFS_MOUNT_TRUNK_DISCOVERY;
+		break;
 	case Opt_extend:
 		if (result.negated)
 			ctx->flags |= NFS_MOUNT_NO_EXTEND;
@@ -943,10 +968,25 @@ static int nfs_fs_context_parse_param(struct fs_context *fc,
 			ctx->flags &= ~NFS_MOUNT_NOACL;
 		break;
 	case Opt_rdirplus:
-		if (result.negated)
+		if (result.negated) {
+			ctx->flags &= ~NFS_MOUNT_FORCE_RDIRPLUS;
 			ctx->flags |= NFS_MOUNT_NORDIRPLUS;
-		else
-			ctx->flags &= ~NFS_MOUNT_NORDIRPLUS;
+		} else if (!param->string) {
+			ctx->flags &= ~(NFS_MOUNT_NORDIRPLUS | NFS_MOUNT_FORCE_RDIRPLUS);
+		} else {
+			switch (lookup_constant(nfs_rdirplus_tokens, param->string, -1)) {
+			case Opt_rdirplus_none:
+				ctx->flags &= ~NFS_MOUNT_FORCE_RDIRPLUS;
+				ctx->flags |= NFS_MOUNT_NORDIRPLUS;
+				break;
+			case Opt_rdirplus_force:
+				ctx->flags &= ~NFS_MOUNT_NORDIRPLUS;
+				ctx->flags |= NFS_MOUNT_FORCE_RDIRPLUS;
+				break;
+			default:
+				goto out_invalid_value;
+			}
+		}
 		break;
 	case Opt_sharecache:
 		if (result.negated)
@@ -977,6 +1017,7 @@ static int nfs_fs_context_parse_param(struct fs_context *fc,
 		ctx->fscache_uniq = NULL;
 		break;
 	case Opt_fscache:
+		trace_nfs_mount_assign(param->key, param->string);
 		ctx->options |= NFS_OPTION_FSCACHE;
 		kfree(ctx->fscache_uniq);
 		ctx->fscache_uniq = param->string;
@@ -990,9 +1031,9 @@ static int nfs_fs_context_parse_param(struct fs_context *fc,
 		break;
 	case Opt_forcerdirplus:
 		if (result.negated)
-			ctx->flags &= ~NFS_MOUNT_FORCE_READDIRPLUS;
+			ctx->flags &= ~NFS_MOUNT_FORCE_RDIRPLUS;
 		else
-			ctx->flags |= NFS_MOUNT_FORCE_READDIRPLUS;
+			ctx->flags |= NFS_MOUNT_FORCE_RDIRPLUS;
 		break;
 	case Opt_relmtime:
 		if (result.negated)
@@ -1002,9 +1043,7 @@ static int nfs_fs_context_parse_param(struct fs_context *fc,
 		break;
 	case Opt_idlexprt:
 		if (result.negated)
-			ctx->no_idle_xprt = 1;
-		else
-			ctx->no_idle_xprt = 0;
+			ctx->no_idle_xprt = true;
 		break;
 	case Opt_mdconnect:
 		if (result.uint_32 < 1 || result.uint_32 > NFS_MAX_MD_CONNECTIONS)
@@ -1086,6 +1125,9 @@ static int nfs_fs_context_parse_param(struct fs_context *fc,
 			return ret;
 		break;
 	case Opt_vers:
+		if (!param->string)
+			goto out_invalid_value;
+		trace_nfs_mount_assign(param->key, param->string);
 		ret = nfs_parse_version_string(fc, param->string);
 		if (ret < 0)
 			return ret;
@@ -1095,8 +1137,16 @@ static int nfs_fs_context_parse_param(struct fs_context *fc,
 		if (ret < 0)
 			return ret;
 		break;
+	case Opt_xprtsec:
+		ret = nfs_parse_xprtsec_policy(fc, param);
+		if (ret < 0)
+			return ret;
+		break;
 
 	case Opt_proto:
+		if (!param->string)
+			goto out_invalid_value;
+		trace_nfs_mount_assign(param->key, param->string);
 		protofamily = AF_INET;
 		switch (lookup_constant(nfs_xprt_protocol_tokens, param->string, -1)) {
 		case Opt_xprt_udp6:
@@ -1132,6 +1182,9 @@ static int nfs_fs_context_parse_param(struct fs_context *fc,
 		break;
 
 	case Opt_mountproto:
+		if (!param->string)
+			goto out_invalid_value;
+		trace_nfs_mount_assign(param->key, param->string);
 		mountfamily = AF_INET;
 		switch (lookup_constant(nfs_xprt_protocol_tokens, param->string, -1)) {
 		case Opt_xprt_udp6:
@@ -1154,6 +1207,7 @@ static int nfs_fs_context_parse_param(struct fs_context *fc,
 		break;
 
 	case Opt_addr:
+		trace_nfs_mount_assign(param->key, param->string);
 		len = rpc_pton(fc->net_ns, param->string, param->size,
 			       &ctx->nfs_server.address,
 			       sizeof(ctx->nfs_server._address));
@@ -1162,16 +1216,19 @@ static int nfs_fs_context_parse_param(struct fs_context *fc,
 		ctx->nfs_server.addrlen = len;
 		break;
 	case Opt_clientaddr:
+		trace_nfs_mount_assign(param->key, param->string);
 		kfree(ctx->client_address);
 		ctx->client_address = param->string;
 		param->string = NULL;
 		break;
 	case Opt_mounthost:
+		trace_nfs_mount_assign(param->key, param->string);
 		kfree(ctx->mount_server.hostname);
 		ctx->mount_server.hostname = param->string;
 		param->string = NULL;
 		break;
 	case Opt_mountaddr:
+		trace_nfs_mount_assign(param->key, param->string);
 		len = rpc_pton(fc->net_ns, param->string, param->size,
 			       &ctx->mount_server.address,
 			       sizeof(ctx->mount_server._address));
@@ -1180,6 +1237,7 @@ static int nfs_fs_context_parse_param(struct fs_context *fc,
 		ctx->mount_server.addrlen = len;
 		break;
 	case Opt_nconnect:
+		trace_nfs_mount_assign(param->key, param->string);
 		if (result.uint_32 < 1 || result.uint_32 > NFS_MAX_CONNECTIONS)
 			goto out_of_bounds;
 		ctx->nfs_server.nconnect = result.uint_32;
@@ -1207,9 +1265,10 @@ static int nfs_fs_context_parse_param(struct fs_context *fc,
 		ret = nfs_parse_port_group(param->string, fc, &ctx->local_ports, "local");
 
 		switch (ret) {
+		case 0: break;
 		case -ENOMEM: goto out_nomem;
 		case -ENOSPC: goto out_portgroup_too_large;
-		case -EINVAL: goto out_invalid_address;
+		default: goto out_invalid_address;
 		}
 		break;
 	case Opt_remoteports:
@@ -1218,26 +1277,20 @@ static int nfs_fs_context_parse_param(struct fs_context *fc,
 		ret = nfs_parse_port_group(param->string, fc, &ctx->remote_ports, "remote");
 
 		switch (ret) {
+		case 0: break;
 		case -ENOMEM: goto out_nomem;
 		case -ENOSPC: goto out_portgroup_too_large;
-		case -EINVAL: goto out_invalid_address;
+		default: goto out_invalid_address;
 		}
 		break;
 	case Opt_max_connect:
+		trace_nfs_mount_assign(param->key, param->string);
 		if (result.uint_32 < 1 || result.uint_32 > NFS_MAX_TRANSPORTS)
 			goto out_of_bounds;
 		ctx->nfs_server.max_connect = result.uint_32;
 		break;
-	case Opt_pconnect:
-		if (result.uint_32 < 1 || result.uint_32 > NFS_MAX_TRANSPORTS)
-			goto out_of_bounds;
-		if (!ctx->nfs_server.nconnect)
-			return nfs_invalf(fc, "NFS: pconnect needs be specified after nconnect");
-		if (result.uint_32 > ctx->nfs_server.nconnect)
-			return nfs_invalf(fc, "NFS: pconnect cannot be larger than nconnect");
-		ctx->nfs_server.pconnect = result.uint_32;
-		break;
 	case Opt_lookupcache:
+		trace_nfs_mount_assign(param->key, param->string);
 		switch (result.uint_32) {
 		case Opt_lookupcache_all:
 			ctx->flags &= ~(NFS_MOUNT_LOOKUP_CACHE_NONEG|NFS_MOUNT_LOOKUP_CACHE_NONE);
@@ -1254,6 +1307,7 @@ static int nfs_fs_context_parse_param(struct fs_context *fc,
 		}
 		break;
 	case Opt_local_lock:
+		trace_nfs_mount_assign(param->key, param->string);
 		switch (result.uint_32) {
 		case Opt_local_lock_all:
 			ctx->flags |= (NFS_MOUNT_LOCAL_FLOCK |
@@ -1274,6 +1328,7 @@ static int nfs_fs_context_parse_param(struct fs_context *fc,
 		}
 		break;
 	case Opt_write:
+		trace_nfs_mount_assign(param->key, param->string);
 		switch (result.uint_32) {
 		case Opt_write_lazy:
 			ctx->flags &=
@@ -1297,10 +1352,18 @@ static int nfs_fs_context_parse_param(struct fs_context *fc,
 		 */
 	case Opt_sloppy:
 		ctx->sloppy = true;
-		dfprintk(MOUNT, "NFS:   relaxing parsing rules\n");
 		break;
 	}
+	{
+		char param_str[0x20] = {0, };
+		const char *value = param_str;
 
+		if (param->type == fs_value_is_string && param->string)
+			value = param->string;
+		else
+			snprintf(param_str, sizeof(param_str), "%d", result.uint_32);
+		trace_nfs_fs_context_parse_param(ctx, param->key, value);
+	}
 	return 0;
 
 out_invalid_value:
@@ -1334,10 +1397,8 @@ static int nfs_parse_source(struct fs_context *fc,
 	size_t len;
 	const char *end;
 
-	if (unlikely(!dev_name || !*dev_name)) {
-		dfprintk(MOUNT, "NFS: device name not specified\n");
+	if (unlikely(!dev_name || !*dev_name))
 		return -EINVAL;
-	}
 
 	/* Is the host name protected with square brakcets? */
 	if (*dev_name == '[') {
@@ -1377,7 +1438,7 @@ static int nfs_parse_source(struct fs_context *fc,
 	if (!ctx->nfs_server.export_path)
 		goto out_nomem;
 
-	dfprintk(MOUNT, "NFS: MNTPATH: '%s'\n", ctx->nfs_server.export_path);
+	trace_nfs_mount_path(ctx->nfs_server.export_path);
 	return 0;
 
 out_bad_devname:
@@ -1419,7 +1480,7 @@ static int nfs23_parse_monolithic(struct fs_context *fc,
 {
 	struct nfs_fs_context *ctx = nfs_fc2context(fc);
 	struct nfs_fh *mntfh = ctx->mntfh;
-	struct sockaddr *sap = (struct sockaddr *)&ctx->nfs_server.address;
+	struct sockaddr_storage *sap = &ctx->nfs_server._address;
 	int extra_flags = NFS_MOUNT_LEGACY_INTERFACE;
 	int ret;
 
@@ -1491,10 +1552,13 @@ static int nfs23_parse_monolithic(struct fs_context *fc,
 		ctx->acdirmax	= data->acdirmax;
 		ctx->need_mount	= false;
 
-		memcpy(sap, &data->addr, sizeof(data->addr));
-		ctx->nfs_server.addrlen = sizeof(data->addr);
-		ctx->nfs_server.port = ntohs(data->addr.sin_port);
-		if (sap->sa_family != AF_INET ||
+		if (!is_remount_fc(fc)) {
+			memcpy(sap, &data->addr, sizeof(data->addr));
+			ctx->nfs_server.addrlen = sizeof(data->addr);
+			ctx->nfs_server.port = ntohs(data->addr.sin_port);
+		}
+
+		if (sap->ss_family != AF_INET ||
 		    !nfs_verify_server_address(sap))
 			goto out_no_address;
 
@@ -1571,7 +1635,6 @@ out_no_sec:
 	return nfs_invalf(fc, "NFS: nfs_mount_data version supports only AUTH_SYS");
 
 out_nomem:
-	dfprintk(MOUNT, "NFS: not enough memory to handle mount options");
 	return -ENOMEM;
 
 out_no_address:
@@ -1651,7 +1714,7 @@ static int nfs4_parse_monolithic(struct fs_context *fc,
 				 struct nfs4_mount_data *data)
 {
 	struct nfs_fs_context *ctx = nfs_fc2context(fc);
-	struct sockaddr *sap = (struct sockaddr *)&ctx->nfs_server.address;
+	struct sockaddr_storage *sap = &ctx->nfs_server._address;
 	int ret;
 	char *c;
 
@@ -1703,7 +1766,7 @@ static int nfs4_parse_monolithic(struct fs_context *fc,
 	if (IS_ERR(c))
 		return PTR_ERR(c);
 	ctx->nfs_server.export_path = c;
-	dfprintk(MOUNT, "NFS: MNTPATH: '%s'\n", c);
+	trace_nfs_mount_path(c);
 
 	c = strndup_user(data->client_addr.data, 16);
 	if (IS_ERR(c))
@@ -1765,7 +1828,7 @@ static int nfs_fs_context_validate(struct fs_context *fc)
 {
 	struct nfs_fs_context *ctx = nfs_fc2context(fc);
 	struct nfs_subversion *nfs_mod;
-	struct sockaddr *sap = (struct sockaddr *)&ctx->nfs_server.address;
+	struct sockaddr_storage *sap = &ctx->nfs_server._address;
 	int max_namelen = PAGE_SIZE;
 	int max_pathlen = NFS_MAXPATHLEN;
 	int port = 0;
@@ -1844,7 +1907,6 @@ static int nfs_fs_context_validate(struct fs_context *fc)
 		if (ret)
 			return ret;
 	}
-
 	/* Load the NFS protocol module if we haven't done so yet */
 	if (!ctx->nfs_mod) {
 		nfs_mod = get_nfs_version(ctx->version);
@@ -2011,11 +2073,9 @@ static int nfs_init_fs_context(struct fs_context *fc)
 		return err;
 	}
 
-	ctx->protofamily		= AF_UNSPEC;
-	ctx->mountfamily		= AF_UNSPEC;
-	ctx->mount_server.port		= NFS_UNSPEC_PORT;
-	ctx->no_idle_xprt		= -1;
-	ctx->local_ports_failover	= true;
+	ctx->protofamily	= AF_UNSPEC;
+	ctx->mountfamily	= AF_UNSPEC;
+	ctx->mount_server.port	= NFS_UNSPEC_PORT;
 
 	if (fc->root) {
 		/* reconfigure, start with the current config */
@@ -2037,7 +2097,7 @@ static int nfs_init_fs_context(struct fs_context *fc)
 		ctx->version		= nfss->nfs_client->rpc_ops->version;
 		ctx->minorversion	= nfss->nfs_client->cl_minorversion;
 
-		memcpy(&ctx->nfs_server.address, &nfss->nfs_client->cl_addr,
+		memcpy(&ctx->nfs_server._address, &nfss->nfs_client->cl_addr,
 			ctx->nfs_server.addrlen);
 
 		if (fc->net_ns != net) {
@@ -2060,17 +2120,18 @@ static int nfs_init_fs_context(struct fs_context *fc)
 		ctx->selected_flavor	= RPC_AUTH_MAXFLAVOR;
 		ctx->minorversion	= 0;
 		ctx->need_mount		= true;
+		ctx->xprtsec.policy	= RPC_XPRTSEC_NONE;
+		ctx->xprtsec.cert_serial	= TLS_NO_CERT;
+		ctx->xprtsec.privkey_serial	= TLS_NO_PRIVKEY;
 
-#if !defined(COMPAT_DETECT_INCLUDE_FS_CONTEXT) || defined(COMPAT_DETECT_FS_CONTEXT_S_IFLAGS)
 		fc->s_iflags		|= SB_I_STABLE_WRITES;
-#endif
 		/*
 		 * Modified defaults
 		 *
 		 * In 2022, the client roundtrip is more expensive than the internal
 		 * server roundrip to bring in the extra data.
 		 */
-		ctx->flags              = NFS_MOUNT_FORCE_READDIRPLUS;
+		ctx->flags              = NFS_MOUNT_FORCE_RDIRPLUS;
 	}
 	fc->fs_private = ctx;
 	fc->ops = &nfs_fs_context_ops;
@@ -2082,17 +2143,11 @@ static int nfs_init_fs_context(struct fs_context *fc)
 struct file_system_type nfs_fs_type = {
 	.owner			= THIS_MODULE,
 	.name			= "nfs",
-#ifdef COMPAT_DETECT_INCLUDE_FS_CONTEXT
 	.init_fs_context	= nfs_init_fs_context,
-	.parameters             = NFS_FS_PARAMETERS_REF,
-#else
-	.mount                  = compat_emulate_fs_context_mount,
-#endif
+	.parameters		= nfs_fs_parameters,
 	.kill_sb		= nfs_kill_super,
 	.fs_flags		= FS_RENAME_DOES_D_MOVE|FS_BINARY_MOUNTDATA,
 };
-COMPAT_FS_CONTEXT_DETECTOR_IMPL(nfs_fs_type, nfs_init_fs_context, NFS_FS_PARAMETERS_REF);
-
 MODULE_ALIAS_FS("nfs");
 EXPORT_SYMBOL_GPL(nfs_fs_type);
 
@@ -2100,17 +2155,11 @@ EXPORT_SYMBOL_GPL(nfs_fs_type);
 struct file_system_type nfs4_fs_type = {
 	.owner			= THIS_MODULE,
 	.name			= "nfs4",
-#ifdef COMPAT_DETECT_INCLUDE_FS_CONTEXT
 	.init_fs_context	= nfs_init_fs_context,
-	.parameters             = NFS_FS_PARAMETERS_REF,
-#else
-	.mount                  = compat_emulate_fs_context_mount,
-#endif
+	.parameters		= nfs_fs_parameters,
 	.kill_sb		= nfs_kill_super,
 	.fs_flags		= FS_RENAME_DOES_D_MOVE|FS_BINARY_MOUNTDATA,
 };
-COMPAT_FS_CONTEXT_DETECTOR_IMPL(nfs4_fs_type, nfs_init_fs_context, NFS_FS_PARAMETERS_REF);
-
 MODULE_ALIAS_FS("nfs4");
 MODULE_ALIAS("nfs4");
 EXPORT_SYMBOL_GPL(nfs4_fs_type);
