@@ -35,14 +35,15 @@
 			_loop_res_ = rpc_call_sync_multipath(clnt, msg, flags, hash, tracecode); \
 			if (_loop_res_ != -EJUKEBOX) \
 				break; \
-			freezable_schedule_timeout_killable_unsafe(NFS_JUKEBOX_RETRY_TIME); \
+			__set_current_state(TASK_KILLABLE|TASK_FREEZABLE_UNSAFE); \
+			schedule_timeout(NFS_JUKEBOX_RETRY_TIME); \
 			_loop_res_ = -ERESTARTSYS; \
 		} while (!fatal_signal_pending(current)); \
 		_loop_res_; \
 	}) \
 
-#define nfs3_rpc_call_sync(clnt, msg, flags, tracecode)	\
-	nfs3_rpc_call_sync_multipath(clnt, msg, flags, 0, tracecode)
+#define nfs3_rpc_call_sync(clnt, msg, flags, tracecode) \
+		nfs3_rpc_call_sync_multipath(clnt, msg, flags, 0, tracecode)
 
 static int
 nfs3_async_handle_jukebox(struct rpc_task *task, struct inode *inode)
@@ -105,8 +106,7 @@ nfs3_proc_get_root(struct nfs_server *server, struct nfs_fh *fhandle,
  */
 static int
 nfs3_proc_getattr(struct nfs_server *server, struct nfs_fh *fhandle,
-		struct nfs_fattr *fattr, struct nfs4_label *label,
-		struct inode *inode)
+		struct nfs_fattr *fattr, struct inode *inode)
 {
 	struct rpc_message msg = {
 		.rpc_proc	= &nfs3_procedures[NFS3PROC_GETATTR],
@@ -123,8 +123,9 @@ nfs3_proc_getattr(struct nfs_server *server, struct nfs_fh *fhandle,
 	dprintk("NFS call  getattr\n");
 	nfs_fattr_init(fattr);
 	status = nfs3_rpc_call_sync_multipath(server->client, &msg, task_flags,
-					 NFS_FH_HASH(fhandle, server->client),
-					 {});
+					      NFS_FH_HASH(fhandle, server->client), {
+		trace_nfs3proc_getattr(_task_, server, fhandle);
+	});
 	dprintk("NFS reply getattr: %d\n", status);
 	return status;
 }
@@ -181,6 +182,7 @@ __nfs3_proc_lookup(struct inode *dir, const char *name, size_t len,
 	};
 	int			status;
 
+	dprintk("%s: call  fsinfo\n", __func__);
 	res.dir_attr = nfs_alloc_fattr();
 	if (res.dir_attr == NULL)
 		return -ENOMEM;
@@ -190,6 +192,8 @@ __nfs3_proc_lookup(struct inode *dir, const char *name, size_t len,
 					      NFS_FH_HASH_INODE(dir), {
 		trace_nfs3proc_lookup(_task_, NFS_SERVER(dir), NFS_FH(dir));
 	});
+	dprintk("%s: reply fsinfo: %d\n", __func__, status);
+
 	nfs_refresh_inode(dir, res.dir_attr);
 	if (status >= 0 && !(fattr->valid & NFS_ATTR_FATTR)) {
 		msg.rpc_proc = &nfs3_procedures[NFS3PROC_GETATTR];
@@ -207,8 +211,7 @@ __nfs3_proc_lookup(struct inode *dir, const char *name, size_t len,
 
 static int
 nfs3_proc_lookup(struct inode *dir, struct dentry *dentry,
-		 struct nfs_fh *fhandle, struct nfs_fattr *fattr,
-		 struct nfs4_label *label)
+		 struct nfs_fh *fhandle, struct nfs_fattr *fattr)
 {
 	unsigned short task_flags = 0;
 
@@ -223,7 +226,7 @@ nfs3_proc_lookup(struct inode *dir, struct dentry *dentry,
 }
 
 static int nfs3_proc_lookupp(struct inode *inode, struct nfs_fh *fhandle,
-			     struct nfs_fattr *fattr, struct nfs4_label *label)
+			     struct nfs_fattr *fattr)
 {
 	const char dotdot[] = "..";
 	const size_t len = strlen(dotdot);
@@ -236,7 +239,8 @@ static int nfs3_proc_lookupp(struct inode *inode, struct nfs_fh *fhandle,
 				  task_flags);
 }
 
-static int nfs3_proc_access(struct inode *inode, struct nfs_access_entry *entry)
+static int nfs3_proc_access(struct inode *inode, struct nfs_access_entry *entry,
+			    const struct cred *cred)
 {
 	struct nfs3_accessargs	arg = {
 		.fh		= NFS_FH(inode),
@@ -247,7 +251,7 @@ static int nfs3_proc_access(struct inode *inode, struct nfs_access_entry *entry)
 		.rpc_proc	= &nfs3_procedures[NFS3PROC_ACCESS],
 		.rpc_argp	= &arg,
 		.rpc_resp	= &res,
-		.rpc_cred	= entry->cred,
+		.rpc_cred	= cred,
 	};
 	int status = -ENOMEM;
 
@@ -334,12 +338,16 @@ nfs3_do_create(struct inode *dir, struct dentry *dentry, struct nfs3_createdata
 {
 	int status;
 
-	status = nfs3_rpc_call_sync_multipath(NFS_CLIENT(dir), &data->msg, 0, NFS_FH_HASH_INODE(dir), {});
+	status = nfs3_rpc_call_sync_multipath(NFS_CLIENT(dir), &data->msg,
+					      0, NFS_FH_HASH_INODE(dir), {
+		trace_nfs3proc_create(_task_, NFS_SERVER(dir), NFS_FH(dir));
+
+	});
 	nfs_post_op_update_inode(dir, data->res.dir_attr);
 	if (status != 0)
 		return ERR_PTR(status);
 
-	return nfs_add_or_obtain(dentry, data->res.fh, data->res.fattr, NULL);
+	return nfs_add_or_obtain(dentry, data->res.fh, data->res.fattr);
 }
 
 static void nfs3_free_createdata(struct nfs3_createdata *data)
@@ -465,7 +473,11 @@ nfs3_proc_remove(struct inode *dir, struct dentry *dentry)
 	if (res.dir_attr == NULL)
 		goto out;
 
-	status = nfs3_rpc_call_sync_multipath(NFS_CLIENT(dir), &msg, 0, NFS_FH_HASH_INODE(dir), {});
+	status = nfs3_rpc_call_sync_multipath(NFS_CLIENT(dir), &msg, 0,
+					      NFS_FH_HASH_INODE(dir), {
+		trace_nfs3proc_remove(_task_, NFS_SERVER(dir), NFS_FH(dir));
+
+	});
 	nfs_post_op_update_inode(dir, res.dir_attr);
 	nfs_free_fattr(res.dir_attr);
 out:
@@ -984,7 +996,7 @@ nfs3_proc_lock(struct file *filp, int cmd, struct file_lock *fl)
 	struct nfs_open_context *ctx = nfs_file_open_context(filp);
 	int status;
 
-	if (fl->fl_flags & FL_CLOSE) {
+	if (*compat_file_lock_flags(fl) & FL_CLOSE) {
 		l_ctx = nfs_get_lock_context(ctx);
 		if (IS_ERR(l_ctx))
 			l_ctx = NULL;
@@ -1000,11 +1012,18 @@ nfs3_proc_lock(struct file *filp, int cmd, struct file_lock *fl)
 	return status;
 }
 
-static int nfs3_have_delegation(struct inode *inode, fmode_t flags)
+static int nfs3_have_delegation(struct inode *inode, fmode_t type, int flags)
 {
 	return 0;
 }
 
+static int nfs3_return_delegation(struct inode *inode)
+{
+	if (S_ISREG(inode->i_mode))
+		nfs_wb_all(inode);
+	return 0;
+}
+
 static const struct inode_operations nfs3_dir_inode_operations = {
 	.create		= nfs_create,
 	.lookup		= nfs_lookup,
@@ -1020,7 +1039,11 @@ static const struct inode_operations nfs3_dir_inode_operations = {
 	.setattr	= nfs_setattr,
 #ifdef CONFIG_NFS_V3_ACL
 	.listxattr	= nfs3_listxattr,
+#ifdef COMPAT_DETECT_GET_INODE_ACL_6_2
+	.get_inode_acl	= nfs3_get_acl,
+#else
 	.get_acl	= nfs3_get_acl,
+#endif
 	.set_acl	= nfs3_set_acl,
 #endif
 };
@@ -1031,7 +1054,11 @@ static const struct inode_operations nfs3_file_inode_operations = {
 	.setattr	= nfs_setattr,
 #ifdef CONFIG_NFS_V3_ACL
 	.listxattr	= nfs3_listxattr,
+#ifdef COMPAT_DETECT_GET_INODE_ACL_6_2
+	.get_inode_acl	= nfs3_get_acl,
+#else
 	.get_acl	= nfs3_get_acl,
+#endif
 	.set_acl	= nfs3_set_acl,
 #endif
 };
@@ -1082,6 +1109,7 @@ const struct nfs_rpc_ops nfs_v3_clientops = {
 	.clear_acl_cache = forget_all_cached_acls,
 	.close_context	= nfs_close_context,
 	.have_delegation = nfs3_have_delegation,
+	.return_delegation = nfs3_return_delegation,
 	.alloc_client	= nfs_alloc_client,
 	.init_client	= nfs_init_client,
 	.free_client	= nfs_free_client,
