@@ -14,10 +14,8 @@ struct nfs_subversion nfs_v3 = {
 	.rpc_vers = &nfs_version3,
 	.rpc_ops  = &nfs_v3_clientops,
 	.sops     = &nfs_sops,
-#if defined(COMPAT_DETECT_POSIX_ACL_ACCESS_XATTR_HANDLER)
-#ifdef CONFIG_NFS_V3_ACL
-       .xattr    = nfs3_xattr_handlers,
-#endif
+#if defined(CONFIG_NFS_V3_ACL) && defined(COMPAT_DETECT_POSIX_ACL_ACCESS_XATTR_HANDLER_6_4)
+        .xattr    = nfs3_xattr_handlers,
 #endif
 };
 
