@@ -19,6 +19,7 @@
 #include "nfs4trace.h"
 #include "pnfs.h"
 #include "netns.h"
+#include "sysfs.h"
 
 #define NFSDBG_FACILITY		NFSDBG_CLIENT
 
@@ -232,9 +233,8 @@ struct nfs_client *nfs4_alloc_client(const struct nfs_client_initdata *cl_init)
 		__set_bit(NFS_CS_INFINITE_SLOTS, &clp->cl_flags);
 	__set_bit(NFS_CS_DISCRTRY, &clp->cl_flags);
 	__set_bit(NFS_CS_NO_RETRANS_TIMEOUT, &clp->cl_flags);
-
-	if (test_bit(NFS_CS_DS, &cl_init->init_flags))
-		__set_bit(NFS_CS_DS, &clp->cl_flags);
+	if (test_bit(NFS_CS_PNFS, &cl_init->init_flags))
+		__set_bit(NFS_CS_PNFS, &clp->cl_flags);
 	/*
 	 * Set up the connection to the server before we add add to the
 	 * global list.
@@ -420,9 +420,6 @@ static void nfs4_add_trunk(struct nfs_client *clp, struct nfs_client *old)
 	};
 	int max_connect = test_bit(NFS_CS_PNFS, &clp->cl_flags) ?
 		clp->cl_max_connect : old->cl_max_connect;
-	struct rpc_clnt_test_and_add_xprt_flags test_and_add_flags = {
-		.in_max_connect = &max_connect,
-	};
 
 	if (clp->cl_proto != old->cl_proto)
 		return;
@@ -436,7 +433,7 @@ static void nfs4_add_trunk(struct nfs_client *clp, struct nfs_client *old)
 	xprt_args.addrlen = clp_salen;
 
 	rpc_clnt_add_xprt(old->cl_rpcclient, &xprt_args,
-			  rpc_clnt_test_and_add_xprt, &test_and_add_flags);
+			  rpc_clnt_test_and_add_xprt, &max_connect);
 }
 
 /**
@@ -904,14 +901,15 @@ nfs4_find_client_sessionid(struct net *net, const struct sockaddr *addr,
  */
 static int nfs4_set_client(struct nfs_server *server,
 		const char *hostname,
-		const struct sockaddr *addr,
+		const struct sockaddr_storage *addr,
 		const size_t addrlen,
 		const char *ip_addr,
 		int proto, const struct rpc_timeout *timeparms,
 		u32 minorversion, unsigned int nconnect,
-		unsigned int max_connect, unsigned int pconnect,
-		struct net *net, unsigned int xprt_id,
-		struct nfs_fs_context *ctx)
+		unsigned int max_connect, struct net *net,
+		unsigned int xprt_id,
+		struct nfs_fs_context *ctx,
+		struct xprtsec_parms *xprtsec)
 {
 	struct nfs_client_initdata cl_init = {
 		.hostname = hostname,
@@ -925,8 +923,8 @@ static int nfs4_set_client(struct nfs_server *server,
 		.timeparms = timeparms,
 		.cred = server->cred,
 		.server_idx = server->idx,
+		.xprtsec = *xprtsec,
 		.xprt_id = xprt_id,
-		.pconnect = pconnect,
 	};
 	struct nfs_client *clp;
 
@@ -937,12 +935,13 @@ static int nfs4_set_client(struct nfs_server *server,
 	switch (proto) {
 	case XPRT_TRANSPORT_RDMA:
 	case XPRT_TRANSPORT_TCP:
+	case XPRT_TRANSPORT_TCP_TLS:
 		cl_init.nconnect = nconnect;
 		break;
 	}
 	if (ctx) {
 		cl_init.local_ports = ctx->local_ports;
-		cl_init.no_idle_xprt = get_no_idle_xprt(ctx);
+		cl_init.no_idle_xprt = ctx->no_idle_xprt;
 		cl_init.local_ports_failover = ctx->local_ports_failover;
 		cl_init.mdconnect = 0; /* Not relevant for NFSv4.x */
 	}
@@ -952,7 +951,7 @@ static int nfs4_set_client(struct nfs_server *server,
 		__set_bit(NFS_CS_MIGRATION, &cl_init.init_flags);
 	if (test_bit(NFS_MIG_TSM_POSSIBLE, &server->mig_status))
 		__set_bit(NFS_CS_TSM_POSSIBLE, &cl_init.init_flags);
-	server->port = rpc_get_port(addr);
+	server->port = rpc_get_port((struct sockaddr *)addr);
 
 	/* Allocate or find a client reference we can use */
 	clp = nfs_get_client(&cl_init);
@@ -974,6 +973,8 @@ static int nfs4_set_client(struct nfs_server *server,
 	set_bit(NFS_CS_CHECK_LEASE_TIME, &clp->cl_res_state);
 
 	server->nfs_client = clp;
+	nfs_sysfs_add_server(server);
+	nfs_sysfs_link_rpc_client(server, clp->cl_rpcclient, "_state");
 	trace_nfs4_server_bind_nfs_client(server);
 
 	return 0;
@@ -990,7 +991,7 @@ static int nfs4_set_client(struct nfs_server *server,
  * the MDS.
  */
 struct nfs_client *nfs4_set_ds_client(struct nfs_server *mds_srv,
-		const struct sockaddr *ds_addr, int ds_addrlen,
+		const struct sockaddr_storage *ds_addr, int ds_addrlen,
 		int ds_proto, unsigned int ds_timeo, unsigned int ds_retrans,
 		u32 minor_version)
 {
@@ -1007,23 +1008,27 @@ struct nfs_client *nfs4_set_ds_client(struct nfs_server *mds_srv,
 		.net = mds_clp->cl_net,
 		.timeparms = &ds_timeout,
 		.cred = mds_srv->cred,
+		.xprtsec = mds_srv->nfs_client->cl_xprtsec,
 		.xprt_id = mds_clp->cl_xprt_id
 	};
 	char buf[INET6_ADDRSTRLEN + 1];
 
-	if (rpc_ntop(ds_addr, buf, sizeof(buf)) <= 0)
+	if (rpc_ntop((struct sockaddr *)ds_addr, buf, sizeof(buf)) <= 0)
 		return ERR_PTR(-EINVAL);
 	cl_init.hostname = buf;
 
-	if (mds_clp->cl_nconnect > 1 && ds_proto == XPRT_TRANSPORT_TCP) {
-		cl_init.nconnect = mds_clp->cl_nconnect;
-		cl_init.max_connect = NFS_MAX_TRANSPORTS;
+	switch (ds_proto) {
+	case XPRT_TRANSPORT_TCP:
+	case XPRT_TRANSPORT_TCP_TLS:
+		if (mds_clp->cl_nconnect > 1) {
+			cl_init.nconnect = mds_clp->cl_nconnect;
+			cl_init.max_connect = NFS_MAX_TRANSPORTS;
+		}
 	}
 
 	if (mds_srv->flags & NFS_MOUNT_NORESVPORT)
 		__set_bit(NFS_CS_NORESVPORT, &cl_init.init_flags);
 
-	__set_bit(NFS_CS_DS, &cl_init.init_flags);
 	__set_bit(NFS_CS_PNFS, &cl_init.init_flags);
 	cl_init.max_connect = NFS_MAX_TRANSPORTS;
 	/*
@@ -1093,31 +1098,15 @@ static void nfs4_session_limit_xasize(struct nfs_server *server)
 #endif
 }
 
-static int nfs4_server_common_setup(struct nfs_server *server,
-		struct nfs_fh *mntfh, bool auth_probe)
+void nfs4_server_set_init_caps(struct nfs_server *server)
 {
-	struct nfs_fattr *fattr;
-	int error;
-
-	/* data servers support only a subset of NFSv4.1 */
-	if (is_ds_only_client(server->nfs_client))
-		return -EPROTONOSUPPORT;
-
-	fattr = nfs_alloc_fattr();
-	if (fattr == NULL)
-		return -ENOMEM;
-
-	/* We must ensure the session is initialised first */
-	error = nfs4_init_session(server->nfs_client);
-	if (error < 0)
-		goto out;
-
 	/* Set the basic capabilities */
 	server->caps |= server->nfs_client->cl_mvops->init_caps;
 	if (server->flags & NFS_MOUNT_NORDIRPLUS)
 			server->caps &= ~NFS_CAP_READDIRPLUS;
 	if (server->nfs_client->cl_proto == XPRT_TRANSPORT_RDMA)
 		server->caps &= ~NFS_CAP_READ_PLUS;
+
 	/*
 	 * Don't use NFS uid/gid mapping if we're using AUTH_SYS or lower
 	 * authentication.
@@ -1125,7 +1114,23 @@ static int nfs4_server_common_setup(struct nfs_server *server,
 	if (nfs4_disable_idmapping &&
 			server->client->cl_auth->au_flavor == RPC_AUTH_UNIX)
 		server->caps |= NFS_CAP_UIDGID_NOMAP;
+}
 
+static int nfs4_server_common_setup(struct nfs_server *server,
+		struct nfs_fh *mntfh, bool auth_probe)
+{
+	int error;
+
+	/* data servers support only a subset of NFSv4.1 */
+	if (is_ds_only_client(server->nfs_client))
+		return -EPROTONOSUPPORT;
+
+	/* We must ensure the session is initialised first */
+	error = nfs4_init_session(server->nfs_client);
+	if (error < 0)
+		goto out;
+
+	nfs4_server_set_init_caps(server);
 
 	/* Probe the root fh to retrieve its FSID and filehandle */
 	error = nfs4_get_rootfh(server, mntfh, auth_probe);
@@ -1137,7 +1142,7 @@ static int nfs4_server_common_setup(struct nfs_server *server,
 			(unsigned long long) server->fsid.minor);
 	nfs_display_fhandle(mntfh, "Pseudo-fs root FH");
 
-	error = nfs_probe_fsinfo(server, mntfh, fattr);
+	error = nfs_probe_server(server, mntfh);
 	if (error < 0)
 		goto out;
 
@@ -1151,7 +1156,6 @@ static int nfs4_server_common_setup(struct nfs_server *server,
 	server->mount_time = jiffies;
 	server->destroy = nfs4_destroy_server;
 out:
-	nfs_free_fattr(fattr);
 	return error;
 }
 
@@ -1183,7 +1187,7 @@ static int nfs4_init_server(struct nfs_server *server, struct fs_context *fc)
 	/* Get a client record */
 	error = nfs4_set_client(server,
 				ctx->nfs_server.hostname,
-				&ctx->nfs_server.address,
+				&ctx->nfs_server._address,
 				ctx->nfs_server.addrlen,
 				ctx->client_address,
 				ctx->nfs_server.protocol,
@@ -1191,15 +1195,15 @@ static int nfs4_init_server(struct nfs_server *server, struct fs_context *fc)
 				ctx->minorversion,
 				ctx->nfs_server.nconnect,
 				ctx->nfs_server.max_connect,
-			 	ctx->nfs_server.pconnect,
-				fc->net_ns, ctx->xprt_id, ctx);
+				fc->net_ns, ctx->xprt_id, ctx,
+				&ctx->xprtsec);
 	if (error < 0)
 		return error;
 
 	if (ctx->rsize)
-		server->rsize = nfs_block_size(ctx->rsize, NULL);
+		server->rsize = nfs_io_size(ctx->rsize, server->nfs_client->cl_proto);
 	if (ctx->wsize)
-		server->wsize = nfs_block_size(ctx->wsize, NULL);
+		server->wsize = nfs_io_size(ctx->wsize, server->nfs_client->cl_proto);
 
 	server->acregmin = ctx->acregmin * HZ;
 	server->acregmax = ctx->acregmax * HZ;
@@ -1254,8 +1258,7 @@ error:
 struct nfs_server *nfs4_create_multi_server(struct fs_context *fc)
 {
 	struct nfs_fs_context *ctx = nfs_fc2context(fc);
-	struct nfs_server *primary = NULL;
-	struct nfs_server *sub;
+	struct nfs_server *primary = NULL, *sub;
 	struct rpc_portgroup *remoteports = ctx->remote_ports;
 	char *orig_hostname = ctx->nfs_server.hostname;
 	u32 orig_nconnect = ctx->nfs_server.nconnect;
@@ -1264,8 +1267,7 @@ struct nfs_server *nfs4_create_multi_server(struct fs_context *fc)
 	bool replace_hostname;
 	uint32_t remoteport_offset = 0, divised_per_client;
 	uint32_t spray_nconnect = orig_nconnect;
-	uint32_t reminder;
-	uint32_t idx = 0;
+	uint32_t reminder, idx = 0;
 	unsigned short port;
 
 	ctx->server_idx = 0;
@@ -1275,77 +1277,40 @@ struct nfs_server *nfs4_create_multi_server(struct fs_context *fc)
 		goto out;
 	}
 	port = rpc_get_port(&ctx->nfs_server.address);
-
-	if (ctx->nfs_server.pconnect > 0) {
-		uint32_t needed_clients;
-		/*
-		 * `pconnect` passed - it specifies the minimum number of connections
-		 * per remote port. We validate and calculate how many remote ports
-		 * to actually use based on nconnect/pconnect ratio.
-		 *
-		 * Examples:
-		 * - nconnect=16,pconnect=4,remoteports=4 -> use 4 ports, 4 conn each
-		 * - nconnect=8,pconnect=4,remoteports=4 -> use 2 ports, 4 conn each
-		 * - nconnect=4,pconnect=4,remoteports=4 -> use 1 port, 4 conn
-		 * - nconnect=16,pconnect=2,remoteports=4 -> use 4 ports, 4 conn each
-		 */
-
-		/* Validate: pconnect cannot exceed nconnect */
-		if (ctx->nfs_server.pconnect > orig_nconnect)
-			return ERR_PTR(-EINVAL);
-
-		needed_clients = (orig_nconnect + ctx->nfs_server.pconnect - 1)
-			/ ctx->nfs_server.pconnect;
-
-		reminder = orig_nconnect % needed_clients;
-		divised_per_client = orig_nconnect / needed_clients;
-	} else {
-		/*
-		 * `pconnect` not passed, we set nconnect per client based on
-		 * amount of remoteports and nconnect.
-		 */
-		reminder = orig_nconnect % remoteports->nr;
-		divised_per_client = orig_nconnect / remoteports->nr;
-	}
+	reminder = orig_nconnect % remoteports->nr;
+	divised_per_client = orig_nconnect / remoteports->nr;
 
 	remoteport_offset = rpc_calc_portgroup_offset(
-		fc->net_ns, 0, remoteports,
-		ctx->nfs_server.remoteports_offset_provided,
-		ctx->nfs_server.remoteports_offset);
-
+				fc->net_ns, 0, remoteports,
+				ctx->nfs_server.remoteports_offset_provided,
+				ctx->nfs_server.remoteports_offset);
 	/*
 	 * If the provided hostname is an IP address, then show
 	 * different IP addresses in /proc/self/net/nfsfs/servers
 	 */
 	replace_hostname = orig_hostname &&
-		rpc_pton(fc->net_ns,
-			 orig_hostname, strlen(orig_hostname),
-			 (struct sockaddr *)&addr, sizeof(addr)) != 0;
+		rpc_pton(fc->net_ns, orig_hostname, strlen(orig_hostname),
+				(struct sockaddr *)&addr, sizeof(addr)) != 0;
 
 	while (spray_nconnect > 0 && idx < NFS_MAX_SERVERS) {
 		char buf[0x100] = {};
 		const int i = remoteport_offset;
 		const uint32_t client_nconnect = divised_per_client
-			+ ((idx < reminder) ? 1 : 0);
-		/*
-		 * One iteration of this loop creates a client with
-		 * `client_nconnect` connections
-		 */
+						+ ((idx < reminder) ? 1 : 0);
 		ctx->nfs_server.nconnect = client_nconnect;
 		ctx->nfs_server.max_connect = divised_per_client + 1;
 
 		rpc_ntop((struct sockaddr *)&remoteports->addrs[i],
-			 buf, sizeof(buf));
-
-		printk(KERN_DEBUG "%s: remoteports[%d] = %s, nconnect=%d\n",
-		       __func__, i, buf, client_nconnect);
+							buf, sizeof(buf));
+		pr_debug("%s: remoteports[%d] = %s, nconnect=%d\n",
+				__func__, i, buf, client_nconnect);
 
 		if (replace_hostname)
 			ctx->nfs_server.hostname = buf;
 
 		ctx->nfs_server.addrlen = remoteports->lens[i];
 		memcpy(&ctx->nfs_server.address, &remoteports->addrs[i],
-		       remoteports->lens[i]);
+							remoteports->lens[i]);
 		rpc_set_port(&ctx->nfs_server.address, port);
 
 		ctx->server_idx = idx;
@@ -1388,8 +1353,8 @@ struct nfs_server *nfs4_create_referral_server(struct fs_context *fc)
 	struct nfs_fs_context *ctx = nfs_fc2context(fc);
 	struct nfs_client *parent_client;
 	struct nfs_server *server, *parent_server;
+	int proto, error;
 	bool auth_probe;
-	int error;
 
 	server = nfs_alloc_server();
 	if (!server)
@@ -1409,7 +1374,7 @@ struct nfs_server *nfs4_create_referral_server(struct fs_context *fc)
 	rpc_set_port(&ctx->nfs_server.address, NFS_RDMA_PORT);
 	error = nfs4_set_client(server,
 				ctx->nfs_server.hostname,
-				&ctx->nfs_server.address,
+				&ctx->nfs_server._address,
 				ctx->nfs_server.addrlen,
 				parent_client->cl_ipaddr,
 				XPRT_TRANSPORT_RDMA,
@@ -1417,27 +1382,30 @@ struct nfs_server *nfs4_create_referral_server(struct fs_context *fc)
 				parent_client->cl_mvops->minor_version,
 				parent_client->cl_nconnect,
 				parent_client->cl_max_connect,
-				parent_client->cl_pconnect,
 				parent_client->cl_net,
-				parent_client->cl_xprt_id, ctx);
+				parent_client->cl_xprt_id, ctx,
+				&parent_client->cl_xprtsec);
 	if (!error)
 		goto init_server;
 #endif	/* IS_ENABLED(CONFIG_SUNRPC_XPRT_RDMA) */
 
+	proto = XPRT_TRANSPORT_TCP;
+	if (parent_client->cl_xprtsec.policy != RPC_XPRTSEC_NONE)
+		proto = XPRT_TRANSPORT_TCP_TLS;
 	rpc_set_port(&ctx->nfs_server.address, NFS_PORT);
 	error = nfs4_set_client(server,
 				ctx->nfs_server.hostname,
-				&ctx->nfs_server.address,
+				&ctx->nfs_server._address,
 				ctx->nfs_server.addrlen,
 				parent_client->cl_ipaddr,
-				XPRT_TRANSPORT_TCP,
+				proto,
 				parent_server->client->cl_timeout,
 				parent_client->cl_mvops->minor_version,
 				parent_client->cl_nconnect,
 				parent_client->cl_max_connect,
-				parent_client->cl_pconnect,
 				parent_client->cl_net,
-				parent_client->cl_xprt_id, ctx);
+				parent_client->cl_xprt_id, ctx,
+				&parent_client->cl_xprtsec);
 	if (error < 0)
 		goto error;
 
@@ -1462,30 +1430,6 @@ error:
 	return ERR_PTR(error);
 }
 
-/*
- * Grab the destination's particulars, including lease expiry time.
- *
- * Returns zero if probe succeeded and retrieved FSID matches the FSID
- * we have cached.
- */
-static int nfs_probe_destination(struct nfs_server *server)
-{
-	struct inode *inode = d_inode(server->super->s_root);
-	struct nfs_fattr *fattr;
-	int error;
-
-	fattr = nfs_alloc_fattr();
-	if (fattr == NULL)
-		return -ENOMEM;
-
-	/* Sanity: the probe won't work if the destination server
-	 * does not recognize the migrated FH. */
-	error = nfs_probe_fsinfo(server, NFS_FH(inode), fattr);
-
-	nfs_free_fattr(fattr);
-	return error;
-}
-
 /**
  * nfs4_update_server - Move an nfs_server to a different nfs_client
  *
@@ -1502,16 +1446,17 @@ static int nfs_probe_destination(struct nfs_server *server)
  * Returns zero on success, or a negative errno value.
  */
 int nfs4_update_server(struct nfs_server *server, const char *hostname,
-		       struct sockaddr *sap, size_t salen, struct net *net)
+		       struct sockaddr_storage *sap, size_t salen, struct net *net)
 {
 	struct nfs_client *clp = server->nfs_client;
 	struct rpc_clnt *clnt = server->client;
 	struct xprt_create xargs = {
 		.ident		= clp->cl_proto,
 		.net		= net,
-		.dstaddr	= sap,
+		.dstaddr	= (struct sockaddr *)sap,
 		.addrlen	= salen,
 		.servername	= hostname,
+		/* cel: bleh. We might need to pass TLS parameters here */
 	};
 	char buf[INET6_ADDRSTRLEN + 1];
 	struct sockaddr_storage address;
@@ -1535,8 +1480,7 @@ int nfs4_update_server(struct nfs_server *server, const char *hostname,
 				clp->cl_proto, clnt->cl_timeout,
 				clp->cl_minorversion,
 				clp->cl_nconnect, clp->cl_max_connect,
-			 	clp->cl_pconnect,
-				net, clp->cl_xprt_id, NULL);
+				net, clp->cl_xprt_id, NULL, &clp->cl_xprtsec);
 	clear_bit(NFS_MIG_TSM_POSSIBLE, &server->mig_status);
 	if (error != 0) {
 		nfs_server_insert_lists(server);
@@ -1551,5 +1495,5 @@ int nfs4_update_server(struct nfs_server *server, const char *hostname,
 	}
 	nfs_server_insert_lists(server);
 
-	return nfs_probe_destination(server);
+	return nfs_probe_server(server, NFS_FH(d_inode(server->super->s_root)));
 }
