@@ -18,10 +18,10 @@
 #include <linux/sunrpc/sched.h>
 #include <linux/sunrpc/xdr.h>
 #include <linux/sunrpc/msg_prot.h>
-#include <linux/sunrpc/debug.h>
 
 #define RPC_MIN_SLOT_TABLE	(2U)
 #define RPC_DEF_SLOT_TABLE	(16U)
+#define RPC_DEF_TCP_MAX_SLOT_TABLE	(128U)
 #define RPC_MAX_SLOT_TABLE_LIMIT	(65536U)
 #define RPC_MAX_SLOT_TABLE	RPC_MAX_SLOT_TABLE_LIMIT
 
@@ -133,6 +133,21 @@ struct rpc_rqst {
 #define rq_svec			rq_snd_buf.head
 #define rq_slen			rq_snd_buf.len
 
+/* RPC transport layer security policies */
+enum xprtsec_policies {
+	RPC_XPRTSEC_NONE = 0,
+	RPC_XPRTSEC_TLS_ANON,
+	RPC_XPRTSEC_TLS_X509,
+};
+
+struct xprtsec_parms {
+	enum xprtsec_policies	policy;
+
+	/* authentication material */
+	key_serial_t		cert_serial;
+	key_serial_t		privkey_serial;
+};
+
 static inline int xprt_rqst_add_seqno(struct rpc_rqst *req, u32 seqno)
 {
 	if (likely(req->rq_seqno_count < RPC_GSS_SEQNO_ARRAY_SIZE))
@@ -160,7 +175,8 @@ struct rpc_xprt_ops {
 	unsigned short	(*get_srcport)(struct rpc_xprt *xprt);
 	int		(*buf_alloc)(struct rpc_task *task);
 	void		(*buf_free)(struct rpc_task *task);
-	void		(*prepare_request)(struct rpc_rqst *req);
+	int		(*prepare_request)(struct rpc_rqst *req,
+					   struct xdr_buf *buf);
 	int		(*send_request)(struct rpc_rqst *req);
 	void		(*wait_for_reply_request)(struct rpc_task *task);
 	void		(*timer)(struct rpc_xprt *xprt, struct rpc_task *task);
@@ -201,6 +217,7 @@ enum xprt_transports {
 	XPRT_TRANSPORT_RDMA	= 256,
 	XPRT_TRANSPORT_BC_RDMA	= XPRT_TRANSPORT_RDMA | XPRT_TRANSPORT_BC,
 	XPRT_TRANSPORT_LOCAL	= 257,
+	XPRT_TRANSPORT_TCP_TLS	= 258,
 };
 
 #define XPRT_MAX_GPU_IDX  64
@@ -250,9 +267,9 @@ struct rpc_xprt {
 	enum rpc_xprt_role_t    role;
 
 #ifdef CONFIG_NVFS
-	char                    pci_devname_address[0x40];
+	char			pci_devname_address[0x40];
 	/* Route priority of this xprt indexed by gpu index */
-	unsigned int            gpu_priority[XPRT_MAX_GPU_IDX];
+	unsigned int		gpu_priority[XPRT_MAX_GPU_IDX];
 #endif
 
 	/*
@@ -260,6 +277,7 @@ struct rpc_xprt {
 	 */
 	unsigned long		bind_timeout,
 				reestablish_timeout;
+	struct xprtsec_parms	xprtsec;
 	unsigned int		connect_cookie;	/* A cookie that gets bumped
 						   every time the transport
 						   is reconnected */
@@ -281,7 +299,7 @@ struct rpc_xprt {
 	spinlock_t		transport_lock;	/* lock transport info */
 	spinlock_t		reserve_lock;	/* lock slot table */
 	spinlock_t		queue_lock;	/* send/receive queue lock */
-	struct rpc_xids	*	xid_space;	/* XID space */
+	struct rpc_xids *	xid_space;	/* XID space */
 	struct rpc_task *	snd_task;	/* Task blocked in send */
 
 	struct list_head	xmit_queue;	/* Send queue */
@@ -319,6 +337,7 @@ struct rpc_xprt {
 	} stat;
 
 	struct net		*xprt_net;
+	netns_tracker		ns_tracker;
 	struct sockaddr_storage localport;
 	size_t			localport_len;
 	struct sockaddr_storage orig_localport;
@@ -395,6 +414,9 @@ struct xprt_create {
 	struct svc_xprt		*bc_xprt;	/* NFSv4.1 backchannel */
 	struct rpc_xprt_switch	*bc_xps;
 	unsigned int		flags;
+	struct xprtsec_parms	xprtsec;
+	unsigned long		connect_timeout;
+	unsigned long		reconnect_timeout;
 };
 
 struct xprt_class {
@@ -402,7 +424,7 @@ struct xprt_class {
 	int			ident;		/* XPRT_TRANSPORT identifier */
 	struct rpc_xprt *	(*setup)(struct xprt_create *);
 #ifdef CONFIG_NVFS
-	bool			(*page_gpu_idx)(struct page *, unsigned int *idx);
+	bool                    (*page_gpu_idx)(struct page *, unsigned int *idx);
 #endif
 	struct module		*owner;
 	char			name[32];
@@ -410,7 +432,7 @@ struct xprt_class {
 };
 
 struct rpc_xids {
-	atomic_t                next;
+	atomic_t		next;
 	struct kref		kref;		/* Reference count */
 };
 
@@ -437,10 +459,9 @@ int			xprt_reserve_xprt_cong(struct rpc_xprt *xprt, struct rpc_task *task);
 void			xprt_alloc_slot(struct rpc_xprt *xprt, struct rpc_task *task);
 void			xprt_free_slot(struct rpc_xprt *xprt,
 				       struct rpc_rqst *req);
-void			xprt_request_prepare(struct rpc_rqst *req);
 bool			xprt_prepare_transmit(struct rpc_task *task);
 void			xprt_request_enqueue_transmit(struct rpc_task *task);
-void			xprt_request_enqueue_receive(struct rpc_task *task);
+int			xprt_request_enqueue_receive(struct rpc_task *task);
 void			xprt_request_wait_receive(struct rpc_task *task);
 void			xprt_request_dequeue_xprt(struct rpc_task *task);
 bool			xprt_request_need_retransmit(struct rpc_task *task);
@@ -461,7 +482,6 @@ bool			xprt_wake_up_backlog(struct rpc_xprt *xprt, struct rpc_rqst *req);
 void			xprt_cleanup_ids(void);
 const char *		xprt_type_to_string(int type);
 
-
 static inline int
 xprt_enable_swap(struct rpc_xprt *xprt)
 {
@@ -508,8 +528,7 @@ bool			xprt_page_gpu_idx(struct page *page, unsigned int *gpu_idx);
 void xprt_portusage_connect(struct xprt_portusage *usage, struct sockaddr *sockaddr,
 			    int *diversion_idx);
 void xprt_portusage_disconnect(struct xprt_portusage *usage, struct sockaddr *sockaddr);
-int xprt_portusage_divert(struct xprt_portusage *usage, struct sockaddr *in_sockaddr,
-			  struct sockaddr_storage *out_sockaddr, size_t *out_sockaddr_len);
+int xprt_portusage_divert(struct rpc_xprt *xprt);
 void xprt_portusage_undivert(struct xprt_portusage *usage, int *idx);
 
 struct rpc_portgroup;
@@ -652,4 +671,7 @@ static inline int xprt_test_and_set_binding(struct rpc_xprt *xprt)
 	return test_and_set_bit(XPRT_BINDING, &xprt->state);
 }
 
+void xprt_set_offline_locked(struct rpc_xprt *xprt, struct rpc_xprt_switch *xps);
+void xprt_set_online_locked(struct rpc_xprt *xprt, struct rpc_xprt_switch *xps);
+void xprt_delete_locked(struct rpc_xprt *xprt, struct rpc_xprt_switch *xps);
 #endif /* _LINUX_SUNRPC_XPRT_H */
