@@ -71,17 +71,16 @@
 /*
  * Local functions
  */
-static void	 xprt_init(struct rpc_xprt *xprt, struct net *net);
+static void	xprt_init(struct rpc_xprt *xprt, struct net *net);
 static __be32	xprt_alloc_xid(struct rpc_xprt *xprt);
-static void	 xprt_destroy(struct rpc_xprt *xprt);
-static void	 xprt_request_init(struct rpc_task *task);
+static void	xprt_destroy(struct rpc_xprt *xprt);
+static void	xprt_request_init(struct rpc_task *task);
+static int	xprt_request_prepare(struct rpc_rqst *req, struct xdr_buf *buf);
 
 #ifdef CONFIG_NVFS
 static bool (*page_gpu_idx)(struct page *, unsigned int *idx);
 #endif
 
-static int	xprt_init_xid(struct rpc_xprt *xprt);
-
 static DEFINE_SPINLOCK(xprt_list_lock);
 static LIST_HEAD(xprt_list);
 
@@ -111,6 +110,10 @@ int xprt_register_transport(struct xprt_class *transport)
 	struct xprt_class *t;
 	int result;
 
+	/* transport doesn't present in kernel */
+	if (!transport || !transport->ident)
+		return -EINVAL;
+
 	result = -EEXIST;
 	spin_lock(&xprt_list_lock);
 	list_for_each_entry(t, &xprt_list, list) {
@@ -264,11 +267,9 @@ EXPORT_SYMBOL_GPL(xprt_find_transport_ident);
 static void xprt_clear_locked(struct rpc_xprt *xprt)
 {
 	xprt->snd_task = NULL;
-	if (!test_bit(XPRT_CLOSE_WAIT, &xprt->state)) {
-		smp_mb__before_atomic();
-		clear_bit(XPRT_LOCKED, &xprt->state);
-		smp_mb__after_atomic();
-	} else
+	if (!test_bit(XPRT_CLOSE_WAIT, &xprt->state))
+		clear_bit_unlock(XPRT_LOCKED, &xprt->state);
+	else
 		queue_work(xprtiod_workqueue, &xprt->task_cleanup);
 }
 
@@ -733,7 +734,7 @@ int xprt_adjust_timeout(struct rpc_rqst *req)
 			req->rq_timeout = to->to_maxval;
 		req->rq_retries++;
 		trace_rpc_task_state_transition(
-			req->rq_task, RPC_TASK_ST_TIMEOUT_RETRY);
+				req->rq_task, RPC_TASK_ST_TIMEOUT_RETRY);
 	} else {
 		req->rq_timeout = to->to_initval;
 		req->rq_retries = 0;
@@ -743,7 +744,7 @@ int xprt_adjust_timeout(struct rpc_rqst *req)
 		rpc_init_rtt(req->rq_task->tk_client->cl_rtt, to->to_initval);
 		spin_unlock(&xprt->transport_lock);
 		trace_rpc_task_state_transition(
-			req->rq_task, RPC_TASK_ST_TIMEOUT_MAJOR);
+				req->rq_task, RPC_TASK_ST_TIMEOUT_MAJOR);
 		status = -ETIMEDOUT;
 	}
 	xprt_reset_minortimeo(req);
@@ -765,6 +766,8 @@ static void xprt_autoclose(struct work_struct *work)
 	unsigned int pflags = memalloc_nofs_save();
 
 	trace_xprt_disconnect_auto(xprt);
+	xprt->connect_cookie++;
+	smp_mb__before_atomic();
 	clear_bit(XPRT_CLOSE_WAIT, &xprt->state);
 	xprt->ops->close(xprt);
 	xprt_release_write(xprt, NULL);
@@ -957,12 +960,7 @@ void xprt_connect(struct rpc_task *task)
 	if (!xprt_lock_write(xprt, task))
 		return;
 
-	if (test_and_clear_bit(XPRT_CLOSE_WAIT, &xprt->state)) {
-		trace_xprt_disconnect_cleanup(xprt);
-		xprt->ops->close(xprt);
-	}
-
-	if (!xprt_connected(xprt)) {
+	if (!xprt_connected(xprt) && !test_bit(XPRT_CLOSE_WAIT, &xprt->state)) {
 		task->tk_rqstp->rq_connect_cookie = xprt->connect_cookie;
 		rpc_sleep_on_timeout(&xprt->pending, task, NULL,
 				xprt_request_timeout(task->tk_rqstp));
@@ -1171,17 +1169,20 @@ xprt_request_need_enqueue_receive(struct rpc_task *task, struct rpc_rqst *req)
  * @task: RPC task
  *
  */
-void
+int
 xprt_request_enqueue_receive(struct rpc_task *task)
 {
 	struct rpc_rqst *req = task->tk_rqstp;
 	struct rpc_xprt *xprt = req->rq_xprt;
+	int ret;
 
 	if (!xprt_request_need_enqueue_receive(task, req))
-		return;
+		return 0;
 
 	trace_rpc_task_state_transition(task, RPC_TASK_ST_QUEUE_RECV);
-	xprt_request_prepare(task->tk_rqstp);
+	ret = xprt_request_prepare(task->tk_rqstp, &req->rq_rcv_buf);
+	if (ret)
+		return ret;
 	spin_lock(&xprt->queue_lock);
 
 	/* Update the softirq receive buffer */
@@ -1195,6 +1196,7 @@ xprt_request_enqueue_receive(struct rpc_task *task)
 
 	/* Turn off autodisconnect */
 	del_timer_sync(&xprt->timer);
+	return 0;
 }
 
 /**
@@ -1247,6 +1249,8 @@ void xprt_complete_rqst(struct rpc_task *task, int copied)
 
 	xprt->stat.recvs++;
 
+	xdr_free_bvec(&req->rq_rcv_buf);
+	req->rq_private_buf.bvec = NULL;
 	req->rq_private_buf.len = copied;
 	/* Ensure all writes are done before we update */
 	/* req->rq_reply_bytes_recvd */
@@ -1365,10 +1369,16 @@ xprt_request_enqueue_transmit(struct rpc_task *task)
 {
 	struct rpc_rqst *pos, *req = task->tk_rqstp;
 	struct rpc_xprt *xprt = req->rq_xprt;
+	int ret;
 
 	trace_rpc_task_state_transition(task, RPC_TASK_ST_READY_QUEUE_XMIT);
 
 	if (xprt_request_need_enqueue_transmit(task, req)) {
+		ret = xprt_request_prepare(task->tk_rqstp, &req->rq_snd_buf);
+		if (ret) {
+			task->tk_status = ret;
+			return;
+		}
 		req->rq_bytes_sent = 0;
 		spin_lock(&xprt->queue_lock);
 		/*
@@ -1435,6 +1445,7 @@ xprt_request_dequeue_transmit_locked(struct rpc_task *task)
 	} else
 		list_del(&req->rq_xmit2);
 	atomic_long_dec(&req->rq_xprt->xmit_queuelen);
+	xdr_free_bvec(&req->rq_snd_buf);
 }
 
 /**
@@ -1471,8 +1482,6 @@ xprt_request_dequeue_xprt(struct rpc_task *task)
 	    test_bit(RPC_TASK_NEED_RECV, &task->tk_runstate) ||
 	    xprt_is_pinned_rqst(req)) {
 		spin_lock(&xprt->queue_lock);
-		xprt_request_dequeue_transmit_locked(task);
-		xprt_request_dequeue_receive_locked(task);
 		while (xprt_is_pinned_rqst(req)) {
 			set_bit(RPC_TASK_MSG_PIN_WAIT, &task->tk_runstate);
 			spin_unlock(&xprt->queue_lock);
@@ -1480,24 +1489,30 @@ xprt_request_dequeue_xprt(struct rpc_task *task)
 			spin_lock(&xprt->queue_lock);
 			clear_bit(RPC_TASK_MSG_PIN_WAIT, &task->tk_runstate);
 		}
+		xprt_request_dequeue_transmit_locked(task);
+		xprt_request_dequeue_receive_locked(task);
 		spin_unlock(&xprt->queue_lock);
+		xdr_free_bvec(&req->rq_rcv_buf);
 	}
 }
 
 /**
  * xprt_request_prepare - prepare an encoded request for transport
  * @req: pointer to rpc_rqst
+ * @buf: pointer to send/rcv xdr_buf
  *
  * Calls into the transport layer to do whatever is needed to prepare
  * the request for transmission or receive.
+ * Returns error, or zero.
  */
-void
-xprt_request_prepare(struct rpc_rqst *req)
+static int
+xprt_request_prepare(struct rpc_rqst *req, struct xdr_buf *buf)
 {
 	struct rpc_xprt *xprt = req->rq_xprt;
 
 	if (xprt->ops->prepare_request)
-		xprt->ops->prepare_request(req);
+		return xprt->ops->prepare_request(req, buf);
+	return 0;
 }
 
 /**
@@ -1525,10 +1540,9 @@ bool xprt_page_gpu_idx(struct page *page, unsigned int *gpu_idx)
 	bool (*fn)(struct page *, unsigned int *) = READ_ONCE(page_gpu_idx);
 	if (fn)
 		return fn(page, gpu_idx);
-
 	return false;
 }
-#endif
+#endif /* CONFIG_NVFS */
 
 /**
  * xprt_prepare_transmit - reserve the transport before sending a request
@@ -1548,6 +1562,9 @@ bool xprt_prepare_transmit(struct rpc_task *task)
 		return false;
 
 	}
+	if (atomic_read(&xprt->swapper))
+		/* This will be clear in __rpc_execute */
+		current->flags |= PF_MEMALLOC;
 	return true;
 }
 
@@ -1734,15 +1751,12 @@ out:
 static struct rpc_rqst *xprt_dynamic_alloc_slot(struct rpc_xprt *xprt)
 {
 	struct rpc_rqst *req = ERR_PTR(-EAGAIN);
-	gfp_t gfp_mask = GFP_KERNEL;
 
 	if (xprt->num_reqs >= xprt->max_reqs)
 		goto out;
 	++xprt->num_reqs;
 	spin_unlock(&xprt->reserve_lock);
-	if (current->flags & PF_WQ_WORKER)
-		gfp_mask |= __GFP_NORETRY | __GFP_NOWARN;
-	req = kzalloc(sizeof(*req), gfp_mask);
+	req = kzalloc(sizeof(*req), rpc_task_gfp_mask());
 	spin_lock(&xprt->reserve_lock);
 	if (req != NULL)
 		goto out;
@@ -1847,7 +1861,7 @@ static int xprt_alloc_id(struct rpc_xprt *xprt)
 {
 	int id;
 
-	id = ida_simple_get(&rpc_xprt_ids, 0, 0, GFP_KERNEL);
+	id = ida_alloc(&rpc_xprt_ids, GFP_KERNEL);
 	if (id < 0)
 		return id;
 
@@ -1857,7 +1871,7 @@ static int xprt_alloc_id(struct rpc_xprt *xprt)
 
 static void xprt_free_id(struct rpc_xprt *xprt)
 {
-	ida_simple_remove(&rpc_xprt_ids, xprt->id);
+	ida_free(&rpc_xprt_ids, xprt->id);
 }
 
 struct rpc_xprt *xprt_alloc(struct net *net, size_t size,
@@ -1866,7 +1880,7 @@ struct rpc_xprt *xprt_alloc(struct net *net, size_t size,
 {
 	struct rpc_xprt *xprt;
 	struct rpc_rqst *req;
-	int i, ret;
+	int i;
 
 	xprt = kzalloc(size, GFP_KERNEL);
 	if (xprt == NULL)
@@ -1875,24 +1889,19 @@ struct rpc_xprt *xprt_alloc(struct net *net, size_t size,
 	xprt_alloc_id(xprt);
 	trace_xprt_alloc(xprt);
 
-	ret = xprt_init_xid(xprt);
-	if (ret) {
-		xprt_free_id(xprt);
-		goto out_free;
-	}
-
 	xprt_init(xprt, net);
 
+	xprt->xid_space = xid_space_alloc();
+	if (!xprt->xid_space)
+		goto out_free;
+
 	for (i = 0; i < num_prealloc; i++) {
 		req = kzalloc(sizeof(struct rpc_rqst), GFP_KERNEL);
 		if (!req)
 			goto out_free;
 		list_add(&req->rq_list, &xprt->free);
 	}
-	if (max_alloc > num_prealloc)
-		xprt->max_reqs = max_alloc;
-	else
-		xprt->max_reqs = num_prealloc;
+	xprt->max_reqs = max_t(unsigned int, max_alloc, num_prealloc);
 	xprt->min_reqs = num_prealloc;
 	xprt->num_reqs = num_prealloc;
 
@@ -1907,7 +1916,7 @@ EXPORT_SYMBOL_GPL(xprt_alloc);
 
 void xprt_free(struct rpc_xprt *xprt)
 {
-	put_net(xprt->xprt_net);
+	put_net_track(xprt->xprt_net, &xprt->ns_tracker);
 	xprt_free_all_slots(xprt);
 	trace_xprt_free(xprt);
 	xprt_free_id(xprt);
@@ -1926,20 +1935,7 @@ xprt_init_connect_cookie(struct rpc_rqst *req, struct rpc_xprt *xprt)
 static __be32
 xprt_alloc_xid(struct rpc_xprt *xprt)
 {
-	return (__force __be32)xid_space_new_xid(xprt->xid_space);
-}
-
-static int
-xprt_init_xid(struct rpc_xprt *xprt)
-{
-	struct rpc_xids	*xids;
-
-	xids = xid_space_alloc();
-	if (IS_ERR(xids))
-		return PTR_ERR(xids);
-
-	xprt->xid_space = xids;
-	return 0;
+	return cpu_to_be32(xid_space_new_xid(xprt->xid_space));
 }
 
 static void
@@ -2044,8 +2040,6 @@ void xprt_release(struct rpc_task *task)
 	spin_unlock(&xprt->transport_lock);
 	if (req->rq_buffer)
 		xprt->ops->buf_free(task);
-	xdr_free_bvec(&req->rq_rcv_buf);
-	xdr_free_bvec(&req->rq_snd_buf);
 	if (req->rq_cred != NULL)
 		put_rpccred(req->rq_cred);
 	if (req->rq_release_snd_buf)
@@ -2104,7 +2098,7 @@ static void xprt_init(struct rpc_xprt *xprt, struct net *net)
 	rpc_init_wait_queue(&xprt->sending, "xprt_sending");
 	rpc_init_priority_wait_queue(&xprt->backlog, "xprt_backlog");
 
-	xprt->xprt_net = get_net(net);
+	xprt->xprt_net = get_net_track(net, &xprt->ns_tracker, GFP_KERNEL);
 }
 
 static void _xprt_portusage_undivert(struct xprt_portusage *usage, int *diversion_idx)
@@ -2164,11 +2158,9 @@ void xprt_portusage_undivert(struct xprt_portusage *usage, int *diversion_idx)
 	spin_unlock(&usage->lock);
 }
 
-int xprt_portusage_divert(struct xprt_portusage *usage,
-			  struct sockaddr *in_sockaddr,
-			  struct sockaddr_storage *out_sockaddr,
-			  size_t *out_sockaddr_len)
+int xprt_portusage_divert(struct rpc_xprt *xprt)
 {
+	struct xprt_portusage *usage = xprt->localport_usage;
 	unsigned int lowest = (unsigned int)(-1);
 	int i, minimum_idx = -1;
 
@@ -2178,7 +2170,8 @@ int xprt_portusage_divert(struct xprt_portusage *usage,
 		struct xprt_port *port = &usage->ports[i];
 		int total_usage;
 
-		if (rpc_cmp_addr_port((struct sockaddr *)&port->addr, in_sockaddr))
+		if (rpc_cmp_addr_port((struct sockaddr *)&port->addr,
+		    (struct sockaddr *)&xprt->orig_localport))
 			continue;
 
 		if (port->xprt_connections == 0)
@@ -2193,8 +2186,9 @@ int xprt_portusage_divert(struct xprt_portusage *usage,
 
 	if (minimum_idx != -1) {
 		struct xprt_port *port = &usage->ports[minimum_idx];
-		*out_sockaddr = port->addr;
-		*out_sockaddr_len = port->len;
+
+		xprt->localport = port->addr;
+		xprt->localport_len = port->len;
 		port->xprt_diversions += 1;
 	}
 
@@ -2245,40 +2239,7 @@ int xprt_portusage_push_address(struct xprt_portusage *usage,
 	return j;
 }
 
-static const struct rpc_call_ops rpc_cb_nconnect_revive_call_ops;
-
-static void rpc_cb_nconnect_revive_done(struct rpc_task *task, void *calldata)
-{
-	struct rpc_rqst *rqst = task->tk_rqstp;
-	struct rpc_xprt *xprt = rqst->rq_xprt;
-
-	dprintk("xprt=%p, revival task done, status=%d",
-		xprt, task->tk_status);
-
-	if (task->tk_status == 0) {
-		xprt_clear_n_recovery(xprt);
-		xprt_clear_n_diversion(xprt);
-	} else {
-		if (RPC_SIGNALLED(task) || kref_read(&xprt->kref) == 0)
-			return;
-
-		schedule_timeout_interruptible(HZ / 2);
-
-		if (RPC_SIGNALLED(task) || kref_read(&xprt->kref) == 0)
-			return;
-
-		task = rpc_call_null_helper(
-			task->tk_client, xprt, NULL,
-			RPC_TASK_ASYNC | RPC_TASK_NULLCREDS,
-			&rpc_cb_nconnect_revive_call_ops, NULL);
-
-		if (!IS_ERR(task)) {
-			dprintk("xprt=%p, recreated revival task\n", xprt);
-			rpc_put_task(task);
-			return;
-		}
-	}
-}
+static void rpc_cb_nconnect_revive_done(struct rpc_task *task, void *calldata);
 
 static void rpc_cb_nconnect_revive_release(void *calldata)
 {
@@ -2289,25 +2250,46 @@ static const struct rpc_call_ops rpc_cb_nconnect_revive_call_ops = {
 	.rpc_release = rpc_cb_nconnect_revive_release,
 };
 
-void
-xprt_diversion_recovery_kickoff(struct rpc_clnt *client, struct rpc_xprt *xprt)
+void rpc_cb_nconnect_revive_done(struct rpc_task *task, void *calldata)
+{
+	struct rpc_clnt *client = task->tk_client;
+	struct rpc_rqst *rqst = task->tk_rqstp;
+	struct rpc_xprt *xprt = rqst->rq_xprt;
+
+	dprintk("xprt=%p, revival task done, status=%d", xprt, task->tk_status);
+
+	if (task->tk_status == 0) {
+		xprt_clear_n_recovery(xprt);
+		xprt_clear_n_diversion(xprt);
+        } else if (!RPC_SIGNALLED(task) && kref_read(&xprt->kref)
+					&& !client->cl_shutdown) {
+		schedule_timeout_interruptible(HZ / 2);
+
+		if (!RPC_SIGNALLED(task))
+			xprt_diversion_recovery_kickoff(client, xprt);
+	}
+}
+
+void xprt_diversion_recovery_kickoff(struct rpc_clnt *client, struct rpc_xprt *xprt)
 {
 	struct rpc_task *task;
 
+	if (!kref_read(&xprt->kref) || client->cl_shutdown)
+		return;
+
 	dprintk("xprt=%p, creating revival task\n", xprt);
 
 	task = rpc_call_null_helper(
-		client, xprt, NULL,
-		RPC_TASK_ASYNC | RPC_TASK_NULLCREDS,
-		&rpc_cb_nconnect_revive_call_ops, NULL);
+			client, xprt, NULL,
+			RPC_TASK_ASYNC | RPC_TASK_NULLCREDS,
+			&rpc_cb_nconnect_revive_call_ops, NULL);
 
-	if (!IS_ERR(task)) {
+	if (IS_ERR_OR_NULL(task)) {
+		dprintk("xprt=%p, ERROR creating revival task\n", xprt);
+	} else {
 		dprintk("xprt=%p, created revival task\n", xprt);
 		rpc_put_task(task);
-		return;
 	}
-
-	dprintk("xprt=%p, ERROR creating revival task\n", xprt);
 }
 
 EXPORT_SYMBOL(xprt_portusage_disconnect);
@@ -2481,16 +2463,45 @@ void xprt_put(struct rpc_xprt *xprt)
 }
 EXPORT_SYMBOL_GPL(xprt_put);
 
-struct rpc_xids *xid_space_alloc(void)
+void xprt_set_offline_locked(struct rpc_xprt *xprt, struct rpc_xprt_switch *xps)
 {
-	struct rpc_xids	*xids;
+	if (!test_and_set_bit(XPRT_OFFLINE, &xprt->state)) {
+		spin_lock(&xps->xps_lock);
+		xps->xps_nactive--;
+		spin_unlock(&xps->xps_lock);
+	}
+}
 
-	xids = kzalloc(sizeof(*xids), GFP_KERNEL);
-	if (xids == NULL)
-		return ERR_PTR(-ENOMEM);
+void xprt_set_online_locked(struct rpc_xprt *xprt, struct rpc_xprt_switch *xps)
+{
+	if (test_and_clear_bit(XPRT_OFFLINE, &xprt->state)) {
+		spin_lock(&xps->xps_lock);
+		xps->xps_nactive++;
+		spin_unlock(&xps->xps_lock);
+	}
+}
 
-	kref_init(&xids->kref);
-	atomic_set(&xids->next, prandom_u32());
+void xprt_delete_locked(struct rpc_xprt *xprt, struct rpc_xprt_switch *xps)
+{
+	if (test_and_set_bit(XPRT_REMOVE, &xprt->state))
+		return;
+
+	xprt_force_disconnect(xprt);
+	if (!test_bit(XPRT_CONNECTED, &xprt->state))
+		return;
+
+	if (!xprt->sending.qlen && !xprt->pending.qlen &&
+	    !xprt->backlog.qlen && !atomic_long_read(&xprt->queuelen))
+		rpc_xprt_switch_remove_xprt(xps, xprt, true);
+}
+
+struct rpc_xids *xid_space_alloc()
+{
+	struct rpc_xids *xids = kzalloc(sizeof(*xids), GFP_KERNEL);
+	if (xids != NULL) {
+		kref_init(&xids->kref);
+		atomic_set(&xids->next, get_random_u32());
+	}
 	return xids;
 }
 
@@ -2512,8 +2523,6 @@ static void xid_space_destroy(struct kref *kref)
 
 void xid_space_put(struct rpc_xids *xids)
 {
-	if (!xids)
-		return;
-
-	kref_put(&xids->kref, xid_space_destroy);
+	if (xids != NULL)
+		kref_put(&xids->kref, xid_space_destroy);
 }
