@@ -29,20 +29,18 @@ typedef struct rpc_xprt *(*xprt_switch_find_xprt_t)(struct rpc_xprt_switch *xps,
 static const struct rpc_xprt_iter_ops rpc_xprt_iter_singular;
 static const struct rpc_xprt_iter_ops rpc_xprt_iter_roundrobin;
 static const struct rpc_xprt_iter_ops rpc_xprt_iter_listall;
+static const struct rpc_xprt_iter_ops rpc_xprt_iter_listoffline;
 
 static void xprt_switch_add_xprt_locked(struct rpc_xprt_switch *xps,
 		struct rpc_xprt *xprt, struct rpc_xids **old_xids)
 {
-	struct rpc_xids	*xids;
 	struct list_head *remoteport_list;
 
 	*old_xids = NULL;
-
 	if (unlikely(xprt_get(xprt) == NULL))
 		return;
 
-	xids = xid_space_get(xps->xps_xid_space);
-	*old_xids = xchg(&xprt->xid_space, xids);
+	*old_xids = xchg(&xprt->xid_space, xid_space_get(xps->xps_xid_space));
 
 	list_add_tail_rcu(&xprt->xprt_switch, &xps->xps_xprt_list);
 	remoteport_list = &xps->xps_remoteport_xprt_list[xprt->remote_port_idx];
@@ -65,7 +63,8 @@ static void xprt_switch_add_xprt_locked(struct rpc_xprt_switch *xps,
 void rpc_xprt_switch_add_xprt(struct rpc_xprt_switch *xps,
 		struct rpc_xprt *xprt)
 {
-	struct rpc_xids	*xids = NULL;
+	struct rpc_xids *xids = NULL;
+
 	if (xprt == NULL)
 		return;
 	spin_lock(&xps->xps_lock);
@@ -79,11 +78,11 @@ void rpc_xprt_switch_add_xprt(struct rpc_xprt_switch *xps,
 }
 
 static void xprt_switch_remove_xprt_locked(struct rpc_xprt_switch *xps,
-		struct rpc_xprt *xprt)
+		struct rpc_xprt *xprt, bool offline)
 {
 	if (unlikely(xprt == NULL))
 		return;
-	if (!test_bit(XPRT_OFFLINE, &xprt->state))
+	if (!test_bit(XPRT_OFFLINE, &xprt->state) && offline)
 		xps->xps_nactive--;
 	xps->xps_nxprts--;
 	if (xps->xps_nxprts == 0)
@@ -97,14 +96,15 @@ static void xprt_switch_remove_xprt_locked(struct rpc_xprt_switch *xps,
  * rpc_xprt_switch_remove_xprt - Removes an rpc_xprt from a rpc_xprt_switch
  * @xps: pointer to struct rpc_xprt_switch
  * @xprt: pointer to struct rpc_xprt
+ * @offline: indicates if the xprt that's being removed is in an offline state
  *
  * Removes xprt from the list of struct rpc_xprt in xps.
  */
 void rpc_xprt_switch_remove_xprt(struct rpc_xprt_switch *xps,
-		struct rpc_xprt *xprt)
+		struct rpc_xprt *xprt, bool offline)
 {
 	spin_lock(&xps->xps_lock);
-	xprt_switch_remove_xprt_locked(xps, xprt);
+	xprt_switch_remove_xprt_locked(xps, xprt, offline);
 	spin_unlock(&xps->xps_lock);
 	trace_xprt_switch_unbind(xps, xprt);
 	xprt_put(xprt);
@@ -121,7 +121,7 @@ static int xprt_switch_alloc_id(struct rpc_xprt_switch *xps, gfp_t gfp_flags)
 {
 	int id;
 
-	id = ida_simple_get(&rpc_xprtswitch_ids, 2000, 0, gfp_flags);
+	id = ida_alloc(&rpc_xprtswitch_ids, gfp_flags);
 	if (id < 0)
 		return id;
 
@@ -131,7 +131,7 @@ static int xprt_switch_alloc_id(struct rpc_xprt_switch *xps, gfp_t gfp_flags)
 
 static void xprt_switch_free_id(struct rpc_xprt_switch *xps)
 {
-	ida_simple_remove(&rpc_xprtswitch_ids, xps->xps_id);
+	ida_free(&rpc_xprtswitch_ids, xps->xps_id);
 }
 
 /**
@@ -146,11 +146,11 @@ struct rpc_xprt_switch *xprt_switch_alloc(struct rpc_xprt *xprt,
 		gfp_t gfp_flags)
 {
 	struct rpc_xprt_switch *xps;
-	struct rpc_xids	*xids;
+	struct rpc_xids *xids;
 	int i;
 
 	xids = xid_space_alloc();
-	if (IS_ERR(xids))
+	if (!xids)
 		return NULL;
 
 	xps = kmalloc(sizeof(*xps), gfp_flags);
@@ -176,7 +176,6 @@ struct rpc_xprt_switch *xprt_switch_alloc(struct rpc_xprt *xprt,
 		trace_xprt_switch_bind(xps, xprt);
 		rpc_sysfs_xprt_setup(xps, xprt, gfp_flags);
 	}
-
 	xid_space_put(xids);
 
 	return xps;
@@ -190,7 +189,7 @@ static void xprt_switch_free_entries(struct rpc_xprt_switch *xps)
 
 		xprt = list_first_entry(&xps->xps_xprt_list,
 				struct rpc_xprt, xprt_switch);
-		xprt_switch_remove_xprt_locked(xps, xprt);
+		xprt_switch_remove_xprt_locked(xps, xprt, true);
 		spin_unlock(&xps->xps_lock);
 		xprt_put(xprt);
 		spin_lock(&xps->xps_lock);
@@ -274,7 +273,6 @@ bool xprt_is_active(const struct rpc_xprt *xprt, struct rpc_xprt_iter_conf conf)
 	if (conf.flags & RPC_XPRT_FLAGS_SKIP_UNCONNECTED)
 		if (xprt_n_diversion(xprt) || xprt_n_recovery(xprt))
 			return false;
-
 #ifdef CONFIG_NVFS
 	if (conf.flags & RPC_XPRT_FLAGS_LOCAL_GPU) {
 		unsigned int priority = xprt->gpu_priority[conf.gpu_idx];
@@ -282,7 +280,6 @@ bool xprt_is_active(const struct rpc_xprt *xprt, struct rpc_xprt_iter_conf conf)
 			return false;
 	}
 #endif
-
 	if (conf.flags & RPC_XPRT_FLAGS_REMOTE_PORT_IDX) {
 		if (xprt->remote_port_idx != conf.remote_port_idx)
 			return false;
@@ -368,6 +365,19 @@ struct rpc_xprt *xprt_switch_find_first_entry(struct rpc_xprt_switch *xps,
 	return NULL;
 }
 
+static
+struct rpc_xprt *xprt_switch_find_first_entry_offline(struct rpc_xprt_switch *xps,
+						      struct rpc_xprt_iter_conf conf)
+{
+	struct rpc_xprt *pos;
+
+	list_for_each_xprt(pos, xps, conf) {
+		if (!xprt_is_active(pos, conf))
+			return pos;
+	}
+	return NULL;
+}
+
 static
 struct rpc_xprt *xprt_iter_first_entry(struct rpc_xprt_iter *xpi,
 				       struct rpc_xprt_iter_conf conf)
@@ -380,9 +390,10 @@ struct rpc_xprt *xprt_iter_first_entry(struct rpc_xprt_iter *xpi,
 }
 
 static
-struct rpc_xprt *xprt_switch_find_current_entry(struct rpc_xprt_switch *xps,
-		const struct rpc_xprt *cur,
-		struct rpc_xprt_iter_conf conf)
+struct rpc_xprt *_xprt_switch_find_current_entry(struct rpc_xprt_switch *xps,
+						 const struct rpc_xprt *cur,
+						 bool find_active,
+						 struct rpc_xprt_iter_conf conf)
 {
 	struct rpc_xprt *pos;
 	bool found = false;
@@ -390,32 +401,69 @@ struct rpc_xprt *xprt_switch_find_current_entry(struct rpc_xprt_switch *xps,
 	list_for_each_xprt(pos, xps, conf) {
 		if (cur == pos)
 			found = true;
-		if (found && xprt_is_active(pos, conf))
+		if (found && ((find_active && xprt_is_active(pos, conf)) ||
+			      (!find_active && !xprt_is_active(pos, conf))))
 			return pos;
 	}
 	return NULL;
 }
 
 static
-struct rpc_xprt *xprt_iter_current_entry(struct rpc_xprt_iter *xpi,
-					 struct rpc_xprt_iter_conf conf)
+struct rpc_xprt *xprt_switch_find_current_entry(struct rpc_xprt_switch *xps,
+						const struct rpc_xprt *cur,
+						struct rpc_xprt_iter_conf conf)
+{
+	return _xprt_switch_find_current_entry(xps, cur, true, conf);
+}
+
+static
+struct rpc_xprt * _xprt_iter_current_entry(struct rpc_xprt_iter *xpi,
+		struct rpc_xprt *first_entry(struct rpc_xprt_switch *xps,
+					     struct rpc_xprt_iter_conf conf),
+		struct rpc_xprt *current_entry(struct rpc_xprt_switch *xps,
+					       const struct rpc_xprt *cur,
+					       struct rpc_xprt_iter_conf conf),
+		struct rpc_xprt_iter_conf conf)
 {
 	struct rpc_xprt_switch *xps = rcu_dereference(xpi->xpi_xpswitch);
-	struct rpc_xprt *xprt;
+	struct list_head *head;
 
 	if (xps == NULL)
 		return NULL;
+	head = &xps->xps_xprt_list;
 	if (xpi->xpi_cursor == NULL || xps->xps_nxprts < 2)
-		return xprt_switch_find_first_entry(xps, conf);
-	xprt = xprt_switch_find_current_entry(xps, xpi->xpi_cursor, conf);
-	if (!xprt && conf.flags != 0)
-		return xprt_switch_find_first_entry(xps, conf);
-
-	return xprt;
+		return first_entry(xps, conf);
+	return current_entry(xps, xpi->xpi_cursor, conf);
 }
 
-bool rpc_xprt_switch_has_addr(struct rpc_xprt_switch *xps,
-			      const struct sockaddr *sap)
+static
+struct rpc_xprt *xprt_iter_current_entry(struct rpc_xprt_iter *xpi,
+					 struct rpc_xprt_iter_conf conf)
+{
+	return _xprt_iter_current_entry(xpi, xprt_switch_find_first_entry,
+			xprt_switch_find_current_entry, conf);
+}
+
+static
+struct rpc_xprt *xprt_switch_find_current_entry_offline(struct rpc_xprt_switch *xps,
+		const struct rpc_xprt *cur, struct rpc_xprt_iter_conf conf)
+{
+	return _xprt_switch_find_current_entry(xps, cur, false, conf);
+}
+
+static
+struct rpc_xprt *xprt_iter_current_entry_offline(struct rpc_xprt_iter *xpi,
+						 struct rpc_xprt_iter_conf conf)
+{
+	return _xprt_iter_current_entry(xpi,
+			xprt_switch_find_first_entry_offline,
+			xprt_switch_find_current_entry_offline,
+			conf);
+}
+
+static
+bool __rpc_xprt_switch_has_addr(struct rpc_xprt_switch *xps,
+				const struct sockaddr *sap)
 {
 	struct rpc_xprt *pos;
 
@@ -432,9 +480,21 @@ bool rpc_xprt_switch_has_addr(struct rpc_xprt_switch *xps,
 	return false;
 }
 
+bool rpc_xprt_switch_has_addr(struct rpc_xprt_switch *xps,
+			      const struct sockaddr *sap)
+{
+	bool res;
+
+	rcu_read_lock();
+	res = __rpc_xprt_switch_has_addr(xps, sap);
+	rcu_read_unlock();
+
+	return res;
+}
+
 static
 struct rpc_xprt *xprt_switch_find_next_entry(struct rpc_xprt_switch *xps,
-		const struct rpc_xprt *cur,
+		const struct rpc_xprt *cur, bool check_active,
 		struct rpc_xprt_iter_conf conf)
 {
 	struct rpc_xprt *pos, *prev = NULL;
@@ -443,7 +503,12 @@ struct rpc_xprt *xprt_switch_find_next_entry(struct rpc_xprt_switch *xps,
 	list_for_each_xprt(pos, xps, conf) {
 		if (cur == prev)
 			found = true;
-		if (found && xprt_is_active(pos, conf))
+		/* for request to return active transports return only
+		 * active, for request to return offline transports
+		 * return only offline
+		 */
+		if (found && ((check_active && xprt_is_active(pos, conf)) ||
+			      (!check_active && !xprt_is_active(pos, conf))))
 			return pos;
 		prev = pos;
 	}
@@ -482,7 +547,7 @@ struct rpc_xprt *__xprt_switch_find_next_entry_roundrobin(struct rpc_xprt_switch
 {
 	struct rpc_xprt *ret;
 
-	ret = xprt_switch_find_next_entry(xps, cur, conf);
+	ret = xprt_switch_find_next_entry(xps, cur, true, conf);
 	if (ret != NULL)
 		return ret;
 	return xprt_switch_find_first_entry(xps, conf);
@@ -532,7 +597,14 @@ struct rpc_xprt *xprt_switch_find_next_entry_all(struct rpc_xprt_switch *xps,
 		const struct rpc_xprt *cur,
 		struct rpc_xprt_iter_conf conf)
 {
-	return xprt_switch_find_next_entry(xps, cur, conf);
+	return xprt_switch_find_next_entry(xps, cur, true, conf);
+}
+
+static
+struct rpc_xprt *xprt_switch_find_next_entry_offline(struct rpc_xprt_switch *xps,
+		const struct rpc_xprt *cur, struct rpc_xprt_iter_conf conf)
+{
+	return xprt_switch_find_next_entry(xps, cur, false, conf);
 }
 
 static
@@ -543,6 +615,14 @@ struct rpc_xprt *xprt_iter_next_entry_all(struct rpc_xprt_iter *xpi,
 			xprt_switch_find_next_entry_all, conf);
 }
 
+static
+struct rpc_xprt *xprt_iter_next_entry_offline(struct rpc_xprt_iter *xpi,
+					      struct rpc_xprt_iter_conf conf)
+{
+	return xprt_iter_next_entry_multiple(xpi,
+			xprt_switch_find_next_entry_offline, conf);
+}
+
 /*
  * xprt_iter_rewind - Resets the xprt iterator
  * @xpi: pointer to rpc_xprt_iter
@@ -550,7 +630,6 @@ struct rpc_xprt *xprt_iter_next_entry_all(struct rpc_xprt_iter *xpi,
  * Resets xpi to ensure that it points to the first entry in the list
  * of transports.
  */
-static
 void xprt_iter_rewind(struct rpc_xprt_iter *xpi)
 {
 	rcu_read_lock();
@@ -596,6 +675,12 @@ void xprt_iter_init_listall(struct rpc_xprt_iter *xpi,
 	__xprt_iter_init(xpi, xps, &rpc_xprt_iter_listall);
 }
 
+void xprt_iter_init_listoffline(struct rpc_xprt_iter *xpi,
+		struct rpc_xprt_switch *xps)
+{
+	__xprt_iter_init(xpi, xps, &rpc_xprt_iter_listoffline);
+}
+
 /**
  * xprt_iter_xchg_switch - Atomically swap out the rpc_xprt_switch
  * @xpi: pointer to rpc_xprt_iter
@@ -731,7 +816,7 @@ bool xprt_find_best_priority(struct rpc_xprt_iter *xpi, unsigned int gpu_idx,
 
 	return ret;
 }
-#endif
+#endif /* CONFIG_NVFS */
 
 /* Policy for always returning the first entry in the rpc_xprt_switch */
 static
@@ -756,3 +841,10 @@ const struct rpc_xprt_iter_ops rpc_xprt_iter_listall = {
 	.xpi_xprt = xprt_iter_current_entry,
 	.xpi_next = xprt_iter_next_entry_all,
 };
+
+static
+const struct rpc_xprt_iter_ops rpc_xprt_iter_listoffline = {
+	.xpi_rewind = xprt_iter_default_rewind,
+	.xpi_xprt = xprt_iter_current_entry_offline,
+	.xpi_next = xprt_iter_next_entry_offline,
+};
