@@ -3,14 +3,15 @@
 
 /* v6.1-rc1~146^2~1 - Change calling conventions for filldir_t */
 
-bool bla_filldir(struct dir_context *d, const char *p, int i, loff_t o, u64 x, unsigned int a)
+static bool bla_filldir(struct dir_context *d, const char *p, int i, loff_t o, u64 x, unsigned int a)
 {
 	return false;
 }
 
-filldir_t check_type(void)
+static filldir_t check_type(void)
 {
 	return bla_filldir;
 }
 
+EXPORT_SYMBOL(check_type);
 MODULE_LICENSE("GPL");
