@@ -993,10 +993,10 @@ void rpc_execute(struct rpc_task *task)
 
 static void rpc_async_schedule(struct work_struct *work)
 {
-	unsigned int pflags = memalloc_nofs_save();
+	unsigned int pflags = memalloc_noio_save();
 
 	__rpc_execute(container_of(work, struct rpc_task, u.tk_work));
-	memalloc_nofs_restore(pflags);
+	memalloc_noio_restore(pflags);
 }
 
 /**
@@ -1160,10 +1160,10 @@ static void rpc_free_task(struct rpc_task *task)
 
 static void rpc_async_release(struct work_struct *work)
 {
-	unsigned int pflags = memalloc_nofs_save();
+	unsigned int pflags = memalloc_noio_save();
 
 	rpc_free_task(container_of(work, struct rpc_task, u.tk_work));
-	memalloc_nofs_restore(pflags);
+	memalloc_noio_restore(pflags);
 }
 
 static void rpc_release_resources_task(struct rpc_task *task)
