@@ -12,6 +12,7 @@
  */
 
 #include <linux/linkage.h>
+#include <linux/sched/signal.h>
 #include <linux/errno.h>
 #include <linux/net.h>
 #include <linux/in.h>
@@ -30,8 +31,6 @@
 
 #include <trace/events/sunrpc.h>
 
-#include "timer.h"
-
 #define RPCDBG_FACILITY	RPCDBG_SVCDSP
 
 static void svc_unregister(const struct svc_serv *serv, struct net *net);
