@@ -21,8 +21,6 @@
 #include <trace/events/sunrpc.h>
 #include <linux/old-kernel.h>
 
-#include "timer.h"
-
 #define RPCDBG_FACILITY	RPCDBG_SVCXPRT
 
 static unsigned int svc_rpc_per_connection_limit __read_mostly;
@@ -771,7 +769,8 @@ static void svc_add_new_temp_xprt(struct svc_serv *serv, struct svc_xprt *newxpt
 	serv->sv_tmpcnt++;
 	if (serv->sv_temptimer.function == NULL) {
 		/* setup timer to age temp transports */
-		timer_set_function(&serv->sv_temptimer, svc_age_temp_xprts);
+		serv->sv_temptimer.function =
+			(TIMER_FUNC_TYPE)svc_age_temp_xprts;
 		mod_timer(&serv->sv_temptimer,
 			  jiffies + svc_conn_age_period * HZ);
 	}
