@@ -47,7 +47,6 @@
 #include <net/checksum.h>
 #include <net/udp.h>
 #include <net/tcp.h>
-#include <linux/bvec.h>
 #include <linux/highmem.h>
 #include <linux/uio.h>
 #include <linux/sched/mm.h>
@@ -337,40 +336,65 @@ xs_alloc_sparse_pages(struct xdr_buf *buf, size_t want, gfp_t gfp)
 	for (i = 0; i < n; i++) {
 		if (buf->pages[i])
 			continue;
-		buf->bvec[i].bv_page = buf->pages[i] = alloc_page(gfp);
+		buf->pages[i] = alloc_page(gfp);
 		if (!buf->pages[i]) {
 			i *= PAGE_SIZE;
 			return i > buf->page_base ? i - buf->page_base : 0;
 		}
+		buf->iovec[i].iov_base = page_address(buf->pages[i]);
 	}
 	return want;
 }
 
 static ssize_t
-xs_sock_recvmsg(struct socket *sock, struct msghdr *msg, int flags, size_t seek)
+xs_sock_recvmsg(struct socket *sock, struct msghdr *msg, int flags, size_t count, size_t seek)
 {
-	ssize_t ret;
-	if (seek != 0)
-		iov_iter_advance(&msg->msg_iter, seek);
-	ret = sock_recvmsg(sock, msg, flags);
+	ssize_t ret, max_count = 0, i, xfer;
+	struct kvec *kvec;
+	struct iov_iter iter;
+	struct iovec *orig = msg->msg_iov;
+	int orig_len = msg->msg_iovlen;
+
+	BUG_ON(seek > count);
+	BUG_ON(orig_len == 0);
+
+	xfer = count - seek;
+	iov_iter_init(&iter, orig, orig_len, xfer, seek);
+	BUG_ON((long)iter.nr_segs < 0);
+
+	kvec = (struct kvec *)iter.iov;
+	kvec->iov_base += iter.iov_offset;
+	kvec->iov_len -= iter.iov_offset;
+
+	for (i = 0; i < iter.nr_segs; i++)
+		max_count += kvec[i].iov_len;
+	if (xfer > max_count && !(flags & MSG_TRUNC))
+		xfer = max_count;
+
+	ret = kernel_recvmsg(sock, msg, kvec, iter.nr_segs, xfer, flags);
+
 	return ret > 0 ? ret + seek : ret;
 }
 
 static ssize_t
 xs_read_kvec(struct socket *sock, struct msghdr *msg, int flags,
-		struct kvec *kvec, size_t count, size_t seek)
+	     struct kvec *kvec, size_t count, size_t seek)
 {
-	iov_iter_kvec(&msg->msg_iter, READ | ITER_KVEC, kvec, 1, count);
-	return xs_sock_recvmsg(sock, msg, flags, seek);
+	struct iovec iovec = {kvec->iov_base, kvec->iov_len};
+	msg->msg_iov = &iovec;
+	msg->msg_iovlen = 1;
+	return xs_sock_recvmsg(sock, msg, flags, count, seek);
 }
 
 static ssize_t
 xs_read_bvec(struct socket *sock, struct msghdr *msg, int flags,
-		struct bio_vec *bvec, unsigned long nr, size_t count,
-		size_t seek)
+	    struct iovec *iovec, struct iovec *tmp_iovec,
+	    unsigned long nr, size_t count, size_t seek)
 {
-	iov_iter_bvec(&msg->msg_iter, READ | ITER_BVEC, bvec, nr, count);
-	return xs_sock_recvmsg(sock, msg, flags, seek);
+	memcpy(tmp_iovec, iovec, nr * sizeof(iovec[0]));
+	msg->msg_iov = tmp_iovec;
+	msg->msg_iovlen = nr;
+	return xs_sock_recvmsg(sock, msg, flags, count, seek);
 }
 
 static ssize_t
@@ -383,20 +407,38 @@ xs_read_discard(struct socket *sock, struct msghdr *msg, int flags,
 
 #if ARCH_IMPLEMENTS_FLUSH_DCACHE_PAGE
 static void
-xs_flush_bvec(const struct bio_vec *bvec, size_t count, size_t seek)
+xs_flush_bvec(const struct iovec *iovec, size_t count, size_t seek)
 {
-	struct bvec_iter bi = {
-		.bi_size = count,
-	};
-	struct bio_vec bv;
+	struct iov_iter iter;
+	unsigned int len;
 
-	bvec_iter_advance(bvec, &bi, seek & PAGE_MASK);
-	for_each_bvec(bv, bvec, bi, bi)
-		flush_dcache_page(bv.bv_page);
+	BUG_ON(seek > count);
+
+	iov_iter_init(&iter, iovec, count - seek, count - seek, seek);
+
+	while (iter.nr_segs > 0 && iter.count > 0) {
+		const bool is_null = iter.iov->iov_base == NULL;
+		const size_t size = min_t(u32,
+			iter.iov->iov_len - iter.iov_offset, iter.count);
+		void *ptr = iter.iov->iov_base + iter.iov_offset;
+		size_t remaining = size;
+
+		while (!is_null && remaining > 0) {
+			flush_dcache_page(virt_to_page(ptr));
+			len = min_t(u32, PAGE_SIZE - offset_in_page(ptr), remaining);
+			ptr += len;
+			remaining -= len;
+		}
+
+		iter.iov++;
+		iter.nr_segs--;
+		iter.count -= size;
+		iter.iov_offset = 0;
+	}
 }
 #else
 static inline void
-xs_flush_bvec(const struct bio_vec *bvec, size_t count, size_t seek)
+xs_flush_bvec(const struct iovec *iovec, size_t count, size_t seek)
 {
 }
 #endif
@@ -428,13 +470,14 @@ xs_read_xdr_buf(struct socket *sock, struct msghdr *msg, int flags,
 			min_t(size_t, count - offset, buf->page_len),
 			GFP_KERNEL);
 	if (seek < want) {
-		ret = xs_read_bvec(sock, msg, flags, buf->bvec,
-				xdr_buf_pagecount(buf),
-				want + buf->page_base,
-				seek + buf->page_base);
+		ret = xs_read_bvec(sock, msg, flags, buf->iovec,
+				   buf->tmp_iovec,
+				   xdr_buf_pagecount(buf),
+				   want + buf->page_base,
+				   seek + buf->page_base);
 		if (ret <= 0)
 			goto sock_err;
-		xs_flush_bvec(buf->bvec, ret, seek + buf->page_base);
+		xs_flush_bvec(buf->iovec, ret, seek + buf->page_base);
 		ret -= buf->page_base;
 		offset += ret;
 		if (offset == count || msg->msg_flags & (MSG_EOR|MSG_TRUNC))
@@ -728,10 +771,10 @@ static void xs_stream_data_receive_workfn(struct work_struct *work)
 {
 	struct sock_xprt *transport =
 		container_of(work, struct sock_xprt, recv_worker);
-	unsigned int pflags = memalloc_nofs_save();
+	unsigned int pflags = memalloc_noio_save();
 
 	xs_stream_data_receive(transport);
-	memalloc_nofs_restore(pflags);
+	memalloc_noio_restore(pflags);
 }
 
 static void
@@ -788,7 +831,7 @@ static int xs_nospace(struct rpc_rqst *req)
 
 		rcu_read_lock();
 		wq = rcu_dereference(sk->sk_wq);
-		set_bit(SOCKWQ_ASYNC_NOSPACE, &wq->flags);
+		set_bit(SOCK_ASYNC_NOSPACE, &sk->sk_socket->flags);
 		rcu_read_unlock();
 
 		sk->sk_write_space(sk);
@@ -1268,14 +1311,15 @@ static void xs_udp_data_read_skb(struct rpc_xprt *xprt,
 	u32 _xid;
 	__be32 *xp;
 
-	repsize = skb->len;
+	repsize = skb->len - sizeof(struct udphdr);
 	if (repsize < 4) {
 		dprintk("RPC:       impossible RPC reply size %d!\n", repsize);
 		return;
 	}
 
 	/* Copy the XID from the skb... */
-	xp = skb_header_pointer(skb, 0, sizeof(_xid), &_xid);
+	xp = skb_header_pointer(skb, sizeof(struct udphdr),
+				sizeof(_xid), &_xid);
 	if (xp == NULL)
 		return;
 
@@ -1295,7 +1339,7 @@ static void xs_udp_data_read_skb(struct rpc_xprt *xprt,
 	/* Suck it into the iovec, verify checksum if not done by hw. */
 	if (csum_partial_copy_to_xdr(&rovr->rq_private_buf, skb)) {
 		spin_lock(&xprt->queue_lock);
-		__UDPX_INC_STATS(sk, UDP_MIB_INERRORS);
+		UDPX_INC_STATS_BH(sk, UDP_MIB_INERRORS);
 		goto out_unpin;
 	}
 
@@ -1305,7 +1349,7 @@ static void xs_udp_data_read_skb(struct rpc_xprt *xprt,
 	spin_unlock(&xprt->transport_lock);
 	spin_lock(&xprt->queue_lock);
 	xprt_complete_rqst(task, copied);
-	__UDPX_INC_STATS(sk, UDP_MIB_INDATAGRAMS);
+	UDPX_INC_STATS_BH(sk, UDP_MIB_INDATAGRAMS);
 out_unpin:
 	xprt_unpin_rqst(rovr);
  out_unlock:
@@ -1339,10 +1383,10 @@ static void xs_udp_data_receive_workfn(struct work_struct *work)
 {
 	struct sock_xprt *transport =
 		container_of(work, struct sock_xprt, recv_worker);
-	unsigned int pflags = memalloc_nofs_save();
+	unsigned int pflags = memalloc_noio_save();
 
 	xs_udp_data_receive(transport);
-	memalloc_nofs_restore(pflags);
+	memalloc_noio_restore(pflags);
 }
 
 /**
@@ -1350,7 +1394,7 @@ static void xs_udp_data_receive_workfn(struct work_struct *work)
  * @sk: socket with data to read
  *
  */
-static void xs_data_ready(struct sock *sk)
+static void xs_data_ready(struct sock *sk, int len)
 {
 	struct rpc_xprt *xprt;
 
@@ -1360,7 +1404,7 @@ static void xs_data_ready(struct sock *sk)
 	if (xprt != NULL) {
 		struct sock_xprt *transport = container_of(xprt,
 				struct sock_xprt, xprt);
-		transport->old_data_ready(sk);
+		transport->old_data_ready(sk, len);
 		/* Any data means we had a useful conversation, so
 		 * then we don't need to delay the next reconnect
 		 */
@@ -1480,7 +1524,7 @@ static void xs_write_space(struct sock *sk)
 	transport = container_of(xprt, struct sock_xprt, xprt);
 	rcu_read_lock();
 	wq = rcu_dereference(sk->sk_wq);
-	if (!wq || test_and_clear_bit(SOCKWQ_ASYNC_NOSPACE, &wq->flags) == 0)
+	if (!wq || test_and_clear_bit(SOCK_ASYNC_NOSPACE, &sk->sk_socket->flags) == 0)
 		goto out;
 
 	xs_run_error_worker(transport, XPRT_SOCK_WAKE_WRITE);
@@ -3160,6 +3204,7 @@ void cleanup_socket_xprt(void)
 	xprt_unregister_transport(&xs_bc_tcp_transport);
 }
 
+#if !defined(COMPAT_DETECT_PARAM_SET_UINT_MINMAX)
 static int param_set_uint_minmax(const char *val,
 		const struct kernel_param *kp,
 		unsigned int min, unsigned int max)
@@ -3177,6 +3222,7 @@ static int param_set_uint_minmax(const char *val,
 	*((unsigned int *)kp->arg) = num;
 	return 0;
 }
+#endif
 
 static int param_set_portnr(const char *val, const struct kernel_param *kp)
 {
