@@ -59,6 +59,9 @@
 #include <linux/version.h>
 #include <linux/module.h>
 
+#include <asm-generic/barrier.h>
+#include <asm/bitops.h>
+
 #include <rdma/ib_cm.h>
 
 #include "xprt_rdma.h"
