@@ -50,7 +50,7 @@
 #include <linux/bvec.h>
 #include <linux/highmem.h>
 #include <linux/uio.h>
-#include <linux/sched/mm.h>
+#include <linux/sched.h>
 
 #include <trace/events/sunrpc.h>
 
@@ -356,7 +356,7 @@ xs_sock_recvmsg(struct socket *sock, struct msghdr *msg, int flags, size_t seek)
 	ssize_t ret;
 	if (seek != 0)
 		iov_iter_advance(&msg->msg_iter, seek);
-	ret = sock_recvmsg(sock, msg, flags);
+	ret = sock_recvmsg(sock, msg, msg_data_left(msg), flags);
 	return ret > 0 ? ret + seek : ret;
 }
 
@@ -732,10 +732,10 @@ static void xs_stream_data_receive_workfn(struct work_struct *work)
 {
 	struct sock_xprt *transport =
 		container_of(work, struct sock_xprt, recv_worker);
-	unsigned int pflags = memalloc_nofs_save();
+	unsigned int pflags = memalloc_noio_save();
 
 	xs_stream_data_receive(transport);
-	memalloc_nofs_restore(pflags);
+	memalloc_noio_restore(pflags);
 }
 
 static void
@@ -792,7 +792,7 @@ static int xs_nospace(struct rpc_rqst *req)
 
 		rcu_read_lock();
 		wq = rcu_dereference(sk->sk_wq);
-		set_bit(SOCK_ASYNC_NOSPACE, &sk->sk_socket->flags);
+		set_bit(SOCKWQ_ASYNC_NOSPACE, &wq->flags);
 		rcu_read_unlock();
 
 		sk->sk_write_space(sk);
@@ -1272,14 +1272,15 @@ static void xs_udp_data_read_skb(struct rpc_xprt *xprt,
 	u32 _xid;
 	__be32 *xp;
 
-	repsize = skb->len;
+	repsize = skb->len - sizeof(struct udphdr);
 	if (repsize < 4) {
 		dprintk("RPC:       impossible RPC reply size %d!\n", repsize);
 		return;
 	}
 
 	/* Copy the XID from the skb... */
-	xp = skb_header_pointer(skb, 0, sizeof(_xid), &_xid);
+	xp = skb_header_pointer(skb, sizeof(struct udphdr),
+				sizeof(_xid), &_xid);
 	if (xp == NULL)
 		return;
 
@@ -1299,7 +1300,7 @@ static void xs_udp_data_read_skb(struct rpc_xprt *xprt,
 	/* Suck it into the iovec, verify checksum if not done by hw. */
 	if (csum_partial_copy_to_xdr(&rovr->rq_private_buf, skb)) {
 		spin_lock(&xprt->queue_lock);
-		__UDPX_INC_STATS(sk, UDP_MIB_INERRORS);
+		UDPX_INC_STATS_BH(sk, UDP_MIB_INERRORS);
 		goto out_unpin;
 	}
 
@@ -1309,7 +1310,7 @@ static void xs_udp_data_read_skb(struct rpc_xprt *xprt,
 	spin_unlock(&xprt->transport_lock);
 	spin_lock(&xprt->queue_lock);
 	xprt_complete_rqst(task, copied);
-	__UDPX_INC_STATS(sk, UDP_MIB_INDATAGRAMS);
+	UDPX_INC_STATS_BH(sk, UDP_MIB_INDATAGRAMS);
 out_unpin:
 	xprt_unpin_rqst(rovr);
  out_unlock:
@@ -1327,7 +1328,7 @@ static void xs_udp_data_receive(struct sock_xprt *transport)
 	if (sk == NULL)
 		goto out;
 	for (;;) {
-		skb = skb_recv_udp(sk, 0, 1, &err);
+		skb = skb_recv_datagram(sk, 0, 1, &err);
 		if (skb == NULL)
 			break;
 		xs_udp_data_read_skb(&transport->xprt, sk, skb);
@@ -1343,10 +1344,10 @@ static void xs_udp_data_receive_workfn(struct work_struct *work)
 {
 	struct sock_xprt *transport =
 		container_of(work, struct sock_xprt, recv_worker);
-	unsigned int pflags = memalloc_nofs_save();
+	unsigned int pflags = memalloc_noio_save();
 
 	xs_udp_data_receive(transport);
-	memalloc_nofs_restore(pflags);
+	memalloc_noio_restore(pflags);
 }
 
 /**
@@ -1484,7 +1485,7 @@ static void xs_write_space(struct sock *sk)
 	transport = container_of(xprt, struct sock_xprt, xprt);
 	rcu_read_lock();
 	wq = rcu_dereference(sk->sk_wq);
-	if (!wq || test_and_clear_bit(SOCK_ASYNC_NOSPACE, &sk->sk_socket->flags) == 0)
+	if (!wq || test_and_clear_bit(SOCKWQ_ASYNC_NOSPACE, &wq->flags) == 0)
 		goto out;
 
 	xs_run_error_worker(transport, XPRT_SOCK_WAKE_WRITE);
